# This file is part of Gajim.
#
# SPDX-License-Identifier: GPL-3.0-only

from __future__ import annotations

import datetime as dt
from zoneinfo import ZoneInfo

from gi.repository import Gtk
from nbxmpp.modules.vcard4 import TzProperty
from nbxmpp.modules.vcard4 import VCard
from nbxmpp.protocol import JID

from gajim.common import app
from gajim.common import types
from gajim.common.i18n import _
from gajim.common.modules.contacts import BareContact

from gajim.gtk.util.misc import get_ui_string


@Gtk.Template(string=get_ui_string("timezone_hint.ui"))
class TimezoneHint(Gtk.Box):
    __gtype_name__ = "TimezoneHint"

    _hint_label: Gtk.Label = Gtk.Template.Child()

    def __init__(self) -> None:
        Gtk.Box.__init__(self)

        self._contact: types.ChatContactT | None = None
        self._remote_timezone: ZoneInfo | None = None

        app.pulse_manager.add_callback(self._update_state)

    def switch_contact(self, contact: types.ChatContactT) -> None:
        self.set_visible(False)

        if self._contact is not None:
            self._contact.disconnect_all_from_obj(self)

        if not isinstance(contact, BareContact):
            self._contact = None
            self._remote_timezone = None
            return

        if contact.subscription != "both":
            return

        self._contact = contact
        self._contact.connect("chatstate-update", self._on_chatstate_update)

        client = app.get_client(contact.account)
        vcard = client.get_module("VCard4").request_vcard(
            jid=self._contact.jid,
            callback=self._on_vcard_received,
            max_cache_seconds=12 * 60 * 60,
        )
        self._process_vcard(vcard)

    def _on_chatstate_update(
        self, _contact: types.ChatContactT, _signal_name: str
    ) -> None:
        # Hide as soon as there is a chat state update (user is present)
        self.set_visible(False)

    def _on_vcard_received(self, jid: JID, vcard: VCard | None) -> None:
        self._process_vcard(vcard)

    def _process_vcard(self, vcard: VCard | None) -> None:
        if vcard is None:
            return

        tz_prop = None
        for prop in vcard.get_properties():
            if isinstance(prop, TzProperty):
                tz_prop = prop

        if tz_prop is None:
            return

        try:
            self._remote_timezone = ZoneInfo(tz_prop.value)
        except Exception:
            return

        self._update_state()

    def _update_state(self) -> None:
        if self._remote_timezone is None:
            return

        time = dt.datetime.now(self._remote_timezone)

        if time.hour > 5:
            # Show hint only between 0:00 and 6:00
            return

        time_string = time.strftime(app.settings.get("time_format"))
        assert self._contact is not None
        self._hint_label.set_markup(
            _("It's <b>%(time)s</b> for %(name)s")
            % {"time": time_string, "name": self._contact.name}
        )
        self.set_visible(True)



# This file is part of Gajim.
#
# SPDX-License-Identifier: GPL-3.0-only

from __future__ import annotations

from typing import Any
from typing import cast
from typing import Literal
from typing import overload

from collections.abc import Callable

from gi.repository import Adw
from gi.repository import GObject
from gi.repository import Gtk

from gajim.common import app
from gajim.common.ged import EventHelper

from gajim.gtk.builder import get_builder
from gajim.gtk.util.classes import SignalManager
from gajim.gtk.window import GajimAppWindow


class Assistant(GajimAppWindow, EventHelper):
    __gsignals__ = {
        "button-clicked": (
            GObject.SignalFlags.RUN_LAST | GObject.SignalFlags.ACTION,
            None,
            (str,),
        ),
        "page-changed": (
            GObject.SignalFlags.RUN_LAST | GObject.SignalFlags.ACTION,
            None,
            (str,),
        ),
    }

    def __init__(
        self,
        name: str = "Assistant",
        transient_for: Gtk.Window | None = None,
        width: int = 550,
        height: int = 400,
        transition_duration: int = 200,
        modal: bool = False,
    ) -> None:
        GajimAppWindow.__init__(
            self,
            name=name,
            transient_for=transient_for,
            default_width=width,
            default_height=height,
            modal=modal,
        )

        EventHelper.__init__(self)

        self._pages: dict[str, AssistantPage] = {}
        self._buttons: dict[str, tuple[Gtk.Button, bool]] = {}
        self._button_visible_func = None
        self._get_current_page_name: str | None = None
        self._last_page_name: str | None = None

        self._ui = get_builder("assistant.ui")
        self._ui.main_box.add_css_class("window-padding")

        toolbar_view = Adw.ToolbarView(content=self._ui.main_box)
        toolbar_view.add_top_bar(Adw.HeaderBar())
        self.set_content(toolbar_view)

        self._ui.stack.set_transition_duration(transition_duration)

        self._connect(
            self._ui.stack, "notify::visible-child-name", self._on_visible_child_name
        )

    def show_first_page(self) -> None:
        page_name = self._ui.stack.get_visible_child_name()
        self.emit("page-changed", page_name)

    def _update_page_complete(self, *args: Any) -> None:
        page_widget = cast(AssistantPage, self._ui.stack.get_visible_child())
        for button, complete in self._buttons.values():
            if complete:
                button.set_sensitive(page_widget.complete)

    def update_title(self) -> None:
        page_widget = cast(AssistantPage, self._ui.stack.get_visible_child())
        self.set_title(page_widget.title)

    def _hide_buttons(self) -> None:
        for button, _complete in self._buttons.values():
            button.set_visible(False)

    def _set_buttons_visible(self) -> None:
        page_name = self._ui.stack.get_visible_child_name()
        assert page_name is not None
        if self._button_visible_func is None:
            buttons = self.get_page(page_name).get_visible_buttons()
            if buttons is not None:
                if len(buttons) == 1:
                    default = buttons[0]
                else:
                    default = self.get_page(page_name).get_default_button()
                if default is not None:
                    self.set_default_button(default)
        else:
            buttons = self._button_visible_func(self, page_name)

        self._update_page_complete()

        if buttons is None:
            return

        for button_name in buttons:
            button, _complete = self._buttons[button_name]
            button.set_visible(True)

    def set_button_visible_func(self, func: Callable[..., list[str]]) -> None:
        self._button_visible_func = func

    def set_default_button(self, button_name: str) -> None:
        button, _complete = self._buttons[button_name]
        self.set_default_widget(button)

    def add_button(
        self,
        name: str,
        label: str,
        css_class: str | None = None,
        complete: bool = False,
    ) -> None:
        button = Gtk.Button(label=label, visible=False)
        self._connect(button, "clicked", self.__on_button_clicked)
        if css_class is not None:
            button.add_css_class(css_class)
        self._buttons[name] = (button, complete)
        self._ui.action_area.append(button)

    def add_pages(self, pages: dict[str, AssistantPage]):
        for name, widget in pages.items():
            self._pages[name] = widget
            self._connect(widget, "update-page-complete", self._update_page_complete)
            self._ui.stack.add_named(widget, name)

    @overload
    def add_default_page(self, name: Literal["success"]) -> AssistantSuccessPage: ...

    @overload
    def add_default_page(self, name: Literal["error"]) -> AssistantErrorPage: ...

    @overload
    def add_default_page(self, name: Literal["progress"]) -> AssistantProgressPage: ...

    def add_default_page(self, name: str) -> AssistantPage:
        if name == "success":
            page = AssistantSuccessPage()
        elif name == "error":
            page = AssistantErrorPage()
        elif name == "progress":
            page = AssistantProgressPage()
        else:
            raise ValueError("Unknown page: %s" % name)

        self._pages[name] = page
        self._ui.stack.add_named(page, name)
        return page

    def get_current_page(self) -> str:
        name = self._ui.stack.get_visible_child_name()
        assert name is not None
        return name

    def show_last_page(self) -> None:
        if self._last_page_name is None:
            return
        self.show_page(self._last_page_name, Gtk.StackTransitionType.SLIDE_RIGHT)

    def show_page(
        self,
        name: str,
        transition: Gtk.StackTransitionType = Gtk.StackTransitionType.NONE,
    ) -> None:
        if self._ui.stack.get_visible_child_name() == name:
            return
        self._hide_buttons()
        self._ui.stack.set_visible_child_full(name, transition)

    def get_page(self, name: str) -> AssistantPage:
        return self._pages[name]

    def _on_visible_child_name(self, stack: Gtk.Stack, _param: str) -> None:
        if stack.get_visible_child_name() is None:
            # Happens for some reason when adding the first page
            return

        self._last_page_name = self._get_current_page_name
        self._get_current_page_name = stack.get_visible_child_name()

        self.update_title()
        self._set_buttons_visible()
        self.emit("page-changed", self._get_current_page_name)

    def __on_button_clicked(self, button: Gtk.Button) -> None:
        for button_name, button_data in self._buttons.items():
            button_ = button_data[0]
            if button_ == button:
                self.emit("button-clicked", button_name)
                return

    def _cleanup(self) -> None:
        self.unregister_events()
        self._pages.clear()
        del self._pages
        self._buttons.clear()
        del self._buttons
        del self._button_visible_func
        self.run_dispose()


class AssistantPage(Gtk.Box, SignalManager):
    __gtype_name__ = "AssistantPage"
    __gsignals__ = {
        "update-page-complete": (GObject.SignalFlags.RUN_LAST, None, ()),
    }

    def __init__(self) -> None:
        Gtk.Box.__init__(
            self,
            orientation=Gtk.Orientation.VERTICAL,
            spacing=18,
            valign=Gtk.Align.CENTER,
        )
        SignalManager.__init__(self)

        self.title: str = ""
        self.complete: bool = True

    def do_unroot(self) -> None:
        Gtk.Box.do_unroot(self)
        self._disconnect_all()
        app.check_finalize(self)

    def get_visible_buttons(self) -> list[str] | None:
        return None

    def get_default_button(self) -> str | None:
        return None

    def set_complete(self, complete: bool) -> None:
        if self.complete == complete:
            return

        self.complete = complete
        self.emit("update-page-complete")

    def update_page_complete(self) -> None:
        self.emit("update-page-complete")


class DefaultPage(AssistantPage):
    __gtype_name__ = "AssistantDefaultPage"

    def __init__(
        self, icon_name: str | None = None, icon_css_class: str | None = None
    ) -> None:
        AssistantPage.__init__(self)

        self._heading = Gtk.Label(
            max_width_chars=30,
            wrap=True,
            halign=Gtk.Align.CENTER,
            justify=Gtk.Justification.CENTER,
        )
        self._heading.add_css_class("title-1")
        self.append(self._heading)

        if icon_name is not None:
            icon = Gtk.Image.new_from_icon_name(icon_name)
            icon.set_pixel_size(64)
            if icon_css_class is not None:
                icon.add_css_class(icon_css_class)
            self.append(icon)

        self._label = Gtk.Label(
            wrap=True,
            max_width_chars=50,
            halign=Gtk.Align.CENTER,
            justify=Gtk.Justification.CENTER,
        )
        self.append(self._label)

    def set_heading(self, heading: str) -> None:
        self._heading.set_text(heading)

    def set_text(self, text: str) -> None:
        self._label.set_text(text)

    def set_title(self, title: str) -> None:
        self.title = title


class AssistantErrorPage(DefaultPage):
    __gtype_name__ = "AssistantErrorPage"

    def __init__(self) -> None:
        DefaultPage.__init__(
            self, icon_name="lucide-circle-x-symbolic", icon_css_class="error"
        )


class AssistantSuccessPage(DefaultPage):
    __gtype_name__ = "AssistantSuccessPage"

    def __init__(self) -> None:
        DefaultPage.__init__(
            self, icon_name="lucide-check-symbolic", icon_css_class="success"
        )


class AssistantProgressPage(AssistantPage):
    __gtype_name__ = "AssistantProgressPage"

    def __init__(self) -> None:
        AssistantPage.__init__(self)

        self._label = Gtk.Label(
            wrap=True,
            max_width_chars=50,
            halign=Gtk.Align.CENTER,
            justify=Gtk.Justification.CENTER,
        )

        spinner = Adw.Spinner()
        self.append(spinner)
        self.append(self._label)

    def set_text(self, text: str) -> None:
        self._label.set_text(text)

    def set_title(self, title: str) -> None:
        self.title = title



from typing import Any
from typing import Literal
from typing import overload

from gi.repository import Adw
from gi.repository import Gtk
from gi.repository import GtkSource

from gajim.common.storage.archive.const import ChatDirection
from gajim.gtk.dropdown import GajimDropDown

class GajimBuilder:
    def __init__(
        self,
        filename: str | None = None,
        instance: Any = None,
        widgets: list[str] | None = None,
        domain: str | None = None,
        gettext_: Any | None = None,
    ) -> None: ...

class Builder(Gtk.Builder): ...

class ActivityChangeTimezoneBuilder(Builder):
    change_timezone_page: Gtk.Box
    old_timezone_label: Gtk.Label
    new_timezone_label: Gtk.Label
    disable_button: Gtk.Button
    update_button: Gtk.Button

class ActivityDefaultBuilder(Builder):
    default_page: Gtk.Box

class ActivityGajimUpdateBuilder(Builder):
    gajim_update_page: Gtk.Box
    update_permission_box: Gtk.Box
    update_permission_title_label: Gtk.Label
    update_permission_enable_button: Gtk.Button
    update_gajim_box: Gtk.Box
    update_gajim_title_label: Gtk.Label
    update_gajim_text_label: Gtk.Label
    update_gajim_button: Gtk.Button
    update_plugins_box: Gtk.Box
    update_plugins_title_label: Gtk.Label
    update_plugins_text_label: Gtk.Label
    update_plugins_automatically_checkbox: Gtk.CheckButton
    update_plugins_show_button: Gtk.Button
    update_plugins_button: Gtk.Button
    update_plugins_success_box: Gtk.Box
    update_plugins_success_title_label: Gtk.Label
    update_plugins_success_text_label: Gtk.Label
    update_plugins_success_button: Gtk.Button

class ActivityMucInvitationBuilder(Builder):
    muc_invitation_page: Gtk.Box
    muc_invitation_box: Gtk.Box
    muc_invitation_declined_box: Gtk.Box
    invitation_image: Gtk.Image
    invitation_title_label: Gtk.Label
    invitation_text_label: Gtk.Label

class ActivityOpenpgpEventBuilder(Builder):
    text_box: Gtk.Box
    title: Gtk.Label
    text: Gtk.Label
    disable_button: Gtk.Button
    backup_button: Gtk.Button
    setup_button: Gtk.Button

class ActivitySubscriptionBuilder(Builder):
    subscription_page: Gtk.Box
    subscription_image: Gtk.Image
    subscription_title_label: Gtk.Label
    subscription_text_label: Gtk.Label
    subscribe_box: Gtk.Box
    subscribe_deny_button: Gtk.Button
    subscribe_accept_button: Gtk.Button
    subscribe_menu_button: Gtk.MenuButton
    unsubscribed_box: Gtk.Box
    unsubscribed_remove_button: Gtk.Button

class AdvancedConfigurationBuilder(Builder):
    box: Gtk.Box
    search_entry: Gtk.SearchEntry
    advanced_treeview: Gtk.TreeView
    treeview_selection: Gtk.TreeSelection
    description: Gtk.Label
    reset_button: Gtk.Button

class AssistantBuilder(Builder):
    main_box: Gtk.Box
    content_area: Gtk.Box
    stack: Gtk.Stack
    action_area: Gtk.Box

class CallWindowBuilder(Builder):
    adjustment1: Gtk.Adjustment
    adjustment2: Gtk.Adjustment
    dtmf_popover: Gtk.Popover
    grid1: Gtk.Grid
    button_1: Gtk.Button
    button_2: Gtk.Button
    button_3: Gtk.Button
    button_4: Gtk.Button
    button_5: Gtk.Button
    button_6: Gtk.Button
    button_7: Gtk.Button
    button_8: Gtk.Button
    button_9: Gtk.Button
    button_star: Gtk.Button
    button_0: Gtk.Button
    button_pound: Gtk.Button
    av_box: Gtk.Box
    video_box: Gtk.Box
    outgoing_viewport: Gtk.Viewport
    incoming_viewport: Gtk.Viewport
    avatar_image: Gtk.Image
    jingle_audio_state: Gtk.Image
    jingle_connection_state: Gtk.Label
    jingle_connection_spinner: Adw.Spinner
    answer_video_button: Gtk.Button
    av_cam_button: Gtk.Button
    av_cam_image: Gtk.Image
    end_call_button: Gtk.Button
    audio_buttons_box: Gtk.Box
    mic_hscale: Gtk.VolumeButton
    volumebutton_plus_button2: Gtk.Button
    volumebutton_minus_button2: Gtk.Button
    sound_hscale: Gtk.VolumeButton
    volumebutton_plus_button1: Gtk.Button
    volumebutton_minus_button1: Gtk.Button
    dtmf_button: Gtk.MenuButton

class ChatControlBuilder(Builder):
    control_box: Gtk.Box
    conv_view_paned: Gtk.Paned
    conv_view_overlay: Gtk.Overlay

class ChatListRowBuilder(Builder):
    mainbox: Gtk.Box
    account_identifier: Gtk.Box
    avatar_image: Gtk.Image
    connection_icon: Gtk.Image
    group_chat_indicator: Gtk.Image
    name_label: Gtk.Label
    chatstate_image: Gtk.Image
    mute_image: Gtk.Image
    timestamp_label: Gtk.Label
    nick_label: Gtk.Label
    message_icon: Gtk.Image
    message_label: Gtk.Label
    mention_indicator: Gtk.Image
    unread_label: Gtk.Label
    revealer: Gtk.Revealer
    close_button: Gtk.Button

class ContactInfoBuilder(Builder):
    groups_model: Gtk.ListStore
    main_stack: Gtk.Stack
    vcard_scrolled: Gtk.ScrolledWindow
    avatar_image: Gtk.Image
    contact_name_controls_box: Gtk.Box
    contact_jid_label: Gtk.Label
    group_chat_info: Adw.PreferencesGroup
    role_info: Adw.ActionRow
    affiliation_info: Adw.ActionRow
    vcard_box: Gtk.Box
    settings_scrolled: Gtk.ScrolledWindow
    settings_box: Gtk.Box
    subscription_box: Gtk.Box
    subscription_listbox: Gtk.ListBox
    from_subscription_switch: Gtk.Switch
    to_subscription_stack: Gtk.Stack
    request_stack: Gtk.Stack
    to_subscription_button: Gtk.Button
    remove_history_button: Gtk.Button
    encryption_omemo_box: Adw.Clamp
    encryption_openpgp_box: Adw.Clamp
    groups_page_stack: Gtk.Stack
    groups_treeview: Gtk.TreeView
    tree_selection: Gtk.TreeSelection
    toggle_renderer: Gtk.CellRendererToggle
    text_renderer: Gtk.CellRendererText
    group_add_button: Gtk.Button
    group_remove_button: Gtk.Button
    notes_page_stack: Gtk.Stack
    scrolledwindow_annotation: Gtk.ScrolledWindow
    textview_annotation: Gtk.TextView
    devices_stack: Gtk.Stack
    devices_page: Adw.PreferencesPage

class CryptoTrustManagerBuilder(Builder):
    qr_code_popover: Gtk.Popover
    comparing_instructions: Gtk.Label
    our_fingerprint_2: Gtk.Label
    qr_code_image: Gtk.Image
    stack: Gtk.Stack
    our_fingerprint_row: Adw.ActionRow
    copy_button: Gtk.Button
    qr_menu_button: Gtk.MenuButton
    manage_trust_button: Gtk.Button
    list_heading: Gtk.Label
    list_heading_box: Gtk.Box
    show_inactive_switch: Gtk.Switch
    remove_public_keys_button: Gtk.Button
    list: Gtk.ListBox
    undecided_placeholder: Gtk.Label

class DbMigrationBuilder(Builder):
    box: Gtk.Box
    stack: Gtk.Stack
    status_label: Gtk.Label
    version_label: Gtk.Label
    error_label: Gtk.Label
    error_view: Gtk.TextView
    error_copy_button: Gtk.Button
    error_close_button: Gtk.Button
    success_close_button: Gtk.Button

class DebugConsoleBuilder(Builder):
    popover: Gtk.Popover
    stanza_presets_listbox: Gtk.ListBox
    stack: Gtk.Stack
    log_view: GtkSource.View
    paned: Gtk.Paned
    search_revealer: Gtk.Revealer
    search_entry: Gtk.SearchEntry
    search_forward: Gtk.Button
    search_backward: Gtk.Button
    search_results_label: Gtk.Label
    scrolled: Gtk.ScrolledWindow
    protocol_view: GtkSource.View
    jump_to_end_button: Gtk.Button
    scrolled_input: Gtk.ScrolledWindow
    input_entry: GtkSource.View
    actionbox: Gtk.Box
    filter_options_button: Gtk.Button
    clear_button: Gtk.Button
    edit_toggle: Gtk.ToggleButton
    paste: Gtk.Button
    menubutton: Gtk.MenuButton
    account_label: Gtk.Label
    send: Gtk.Button
    header_box: Gtk.Box
    search_toggle: Gtk.ToggleButton

class ExceptionDialogBuilder(Builder):
    exception_box: Gtk.Box
    infobar: Gtk.Revealer
    exception_view: Gtk.TextView
    user_feedback_box: Gtk.Box
    user_feedback_entry: Gtk.Entry
    close_button: Gtk.Button
    report_spinner: Adw.Spinner
    report_button: Gtk.Button

class FileTransferBuilder(Builder):
    transfer_box: Gtk.Box
    transfer_description: Gtk.Label
    file_name: Gtk.Label
    file_size: Gtk.Label
    progress_bar: Gtk.ProgressBar
    transfer_progress: Gtk.Label
    cancel_button: Gtk.Button

class FileTransferJingleBuilder(Builder):
    transfer_box: Gtk.Box
    transfer_action: Gtk.Label
    file_name: Gtk.Label
    file_description: Gtk.Label
    file_size: Gtk.Label
    action_stack: Gtk.Stack
    accept_file_request: Gtk.Button
    reject_file_request: Gtk.Button
    open_folder: Gtk.Button
    open_file: Gtk.Button
    error_show_transfers: Gtk.Button
    error_label: Gtk.Label
    retry_bad_hash: Gtk.Button
    rejected_show_transfers: Gtk.Button
    progress_label: Gtk.Label
    progress_bar: Gtk.ProgressBar
    cancel_transfer: Gtk.Button

class FileTransferSelectorBuilder(Builder):
    file_box: Gtk.Box
    preview_image_box: Gtk.Box
    preview_image: Gtk.Image
    file_name_label: Gtk.Label
    file_size_label: Gtk.Label
    warning_label: Gtk.Label
    remove_file_button: Gtk.Button
    stack: Gtk.Stack
    box: Gtk.Box
    listbox: Gtk.ListBox
    resource_box: Gtk.Box
    resource_instructions: Gtk.Label

class GroupchatAffiliationBuilder(Builder):
    affiliation_store: Gtk.ListStore
    combo_store: Gtk.ListStore
    main_box: Gtk.Box
    affiliation_scrolled: Gtk.ScrolledWindow
    affiliation_treeview: Gtk.TreeView
    affiliation_selection: Gtk.TreeSelection
    address_renderer: Gtk.CellRendererText
    reserved_name_column: Gtk.TreeViewColumn
    reserved_name_renderer: Gtk.CellRendererText
    affiliation_renderer: Gtk.CellRendererCombo
    button_box: Gtk.Box
    add_remove_button_box: Gtk.Box
    add_button: Gtk.Button
    remove_button: Gtk.Button

class GroupchatBlocksBuilder(Builder):
    main: Gtk.Box
    top_box: Gtk.Box
    search_entry: Gtk.SearchEntry
    remove_button: Gtk.Button
    scrolled_box: Gtk.Box
    scrolled: Gtk.ScrolledWindow
    column_view: Gtk.ColumnView
    nickname_col: Gtk.ColumnViewColumn
    id_col: Gtk.ColumnViewColumn

class GroupchatConfigBuilder(Builder):
    stack: Gtk.Stack
    loading_box: Gtk.Box
    config_box: Gtk.Box
    error_box: Gtk.Box
    error_image: Gtk.Image
    error_label: Gtk.Label

class GroupchatDetailsBuilder(Builder):
    main_stack: Gtk.Stack
    info_container: Gtk.Box
    settings_box: Gtk.Box
    encryption_omemo_box: Adw.Clamp
    encryption_openpgp_box: Adw.Clamp
    blocks_box: Gtk.Box
    manage_box: Gtk.Box
    affiliation_box: Gtk.Box
    outcasts_box: Gtk.Box
    configuration_box: Gtk.Box

class GroupchatInfoScrolledBuilder(Builder):
    info_clamp: Adw.Clamp
    avatar_image: Gtk.Image
    name_box: Gtk.Box
    info_listbox: Gtk.ListBox
    address_row: Adw.ActionRow
    address_copy_button: Gtk.Button
    description_row: Adw.ActionRow
    subject_row: Adw.ActionRow
    author_row: Adw.ActionRow
    users_row: Adw.ActionRow
    contact_row: Adw.ActionRow
    contact_box: Gtk.Box
    logs_row: Adw.ActionRow
    logs: Gtk.LinkButton
    lang_row: Adw.ActionRow
    features_group: Adw.PreferencesGroup
    features_listbox: Gtk.ListBox

class GroupchatInviterBuilder(Builder):
    account_store: Gtk.ListStore
    invite_box: Gtk.Box
    search_entry: Gtk.SearchEntry
    scrolledwindow: Gtk.ScrolledWindow
    contacts_listbox: Gtk.ListBox
    contacts_placeholder: Gtk.Box
    invitees_scrolled: Gtk.ScrolledWindow

class GroupchatManageBuilder(Builder):
    scrolled: Gtk.ScrolledWindow
    stack: Gtk.Stack
    avatar_overlay: Gtk.Overlay
    avatar_button_image: Gtk.Image
    remove_avatar_button: Gtk.Button
    manage_save_button: Gtk.Button
    contact_addresses_listbox: Gtk.ListBox
    muc_name_entry_row: Adw.EntryRow
    muc_description_entry_row: Adw.EntryRow
    destroy_muc_button: Gtk.Button
    subject_change_button: Gtk.Button
    subject_textview: Gtk.TextView
    avatar_selector_grid: Gtk.Grid
    avatar_cancel_button: Gtk.Button
    avatar_update_button: Gtk.Button
    destroy_reason_entry: Gtk.Entry
    destroy_alternate_entry: Gtk.Entry
    destroy_cancel_button: Gtk.Button
    destroy_button: Gtk.Button

class GroupchatOutcastBuilder(Builder):
    info_popover: Gtk.Popover
    outcast_store: Gtk.ListStore
    main_box: Gtk.Box
    outcast_scrolled: Gtk.ScrolledWindow
    outcast_treeview: Gtk.TreeView
    outcast_selection: Gtk.TreeSelection
    address_renderer: Gtk.CellRendererText
    reason_renderer: Gtk.CellRendererText
    button_box: Gtk.Box
    add_remove_button_box: Gtk.Box
    add_button: Gtk.Button
    remove_button: Gtk.Button
    info_button: Gtk.MenuButton

class GroupchatRosterBuilder(Builder):
    box: Gtk.Box
    participants_count_label: Gtk.Label
    search_entry: Gtk.SearchEntry
    scrolled: Gtk.ScrolledWindow

class GroupchatRosterTooltipBuilder(Builder):
    tooltip_grid: Gtk.Grid
    avatar: Gtk.Image
    jid: Gtk.Label
    nick: Gtk.Label
    fillelement: Gtk.Label
    status: Gtk.Label
    affiliation: Gtk.Label
    hats_box: Gtk.Box

class GroupchatStateBuilder(Builder):
    groupchat_state: Gtk.Stack
    join_button: Gtk.Button
    joining_spinner: Adw.Spinner
    abort_join_button: Gtk.Button
    mam_sync_spinner: Adw.Spinner
    mam_error_label: Gtk.Label
    close_button: Gtk.Button

class HistoryExportBuilder(Builder):
    select_account_box: Gtk.Box
    settings_grid: Gtk.Grid

class ManageSoundsBuilder(Builder):
    manage_sounds: Gtk.Box
    sounds_treeview: Gtk.TreeView
    toggle_cell_renderer: Gtk.CellRendererToggle
    sound_buttons_box: Gtk.Box
    clear_sound_button: Gtk.Button
    play_sound_button: Gtk.Button

class MessageActionsBoxBuilder(Builder):
    box: Gtk.Box
    hint_box: Gtk.Box
    link_previews_box: Gtk.Box
    reply_box: Gtk.Box
    state_box: Gtk.Box
    state_box_image: Gtk.Image
    state_box_label: Gtk.Label
    visitor_menu_button: Gtk.MenuButton
    edit_box: Gtk.Box
    edit_box_image: Gtk.Image
    cancel_correction_button: Gtk.Button
    action_box: Gtk.Box
    emoticons_button: Gtk.MenuButton
    formattings_button: Gtk.MenuButton
    input_wrapper: Gtk.Overlay
    input_scrolled: Gtk.ScrolledWindow
    input_overlay: Gtk.Box
    input_overlay_label: Gtk.Label
    send_message_button: Gtk.Button
    sendfile_button: Gtk.Button
    encryption_menu_button: Gtk.MenuButton
    encryption_image: Gtk.Image
    encryption_details_button: Gtk.Button
    encryption_details_image: Gtk.Image
    overflow_menu_button: Gtk.MenuButton
    overflow_popover: Gtk.Popover
    overflow_box: Gtk.Box
    visitor_popover: Gtk.Popover
    request_voice_button: Gtk.Button

class PasswordDialogBuilder(Builder):
    pass_box: Gtk.Box
    header: Gtk.Label
    message_label: Gtk.Label
    pass_entry: Gtk.PasswordEntry
    save_pass_checkbutton: Gtk.CheckButton
    keyring_hint: Gtk.Label
    cancel_button: Gtk.Button
    ok_button: Gtk.Button

class PepConfigBuilder(Builder):
    stack: Gtk.Stack
    overview_box: Gtk.Box
    services_treeview: Gtk.TreeView
    reload_button: Gtk.Button
    delete_button: Gtk.Button
    configure_button: Gtk.Button
    show_content_button: Gtk.Button
    items_box: Gtk.Box
    items_label: Gtk.Label
    items_view: GtkSource.View
    items_back_button: Gtk.Button
    config_box: Gtk.Box
    form_label: Gtk.Label
    form_box: Gtk.Box
    config_back_button: Gtk.Button
    save_button: Gtk.Button

class ProfileBuilder(Builder):
    privacy_popover: Gtk.Popover
    avatar_nick_access: Gtk.Switch
    vcard_access: Gtk.Switch
    avatar_nick_access_label: Gtk.Label
    vcard_access_label: Gtk.Label
    profile_stack: Gtk.Stack
    spinner: Adw.Spinner
    scrolled: Gtk.ScrolledWindow
    profile_box: Gtk.Box
    avatar_overlay: Gtk.Overlay
    avatar_image: Gtk.Image
    remove_avatar_button: Gtk.Button
    nickname_entry: Gtk.Entry
    nickname_label: Gtk.Label
    cancel_button: Gtk.Button
    add_entry_button: Gtk.MenuButton
    privacy_button: Gtk.MenuButton
    save_button: Gtk.Button
    edit_button: Gtk.Button
    avatar_selector_box: Gtk.Box
    avatar_cancel: Gtk.Button
    avatar_update_button: Gtk.Button
    error_label: Gtk.Label
    error_title_label: Gtk.Label
    back_button: Gtk.Button

class QuitDialogBuilder(Builder):
    box: Gtk.Box
    remember_checkbutton: Gtk.CheckButton
    hide_button: Gtk.Button
    minimize_button: Gtk.Button
    quit_button: Gtk.Button

class RosterItemExchangeBuilder(Builder):
    roster_item_exchange: Gtk.Box
    type_label: Gtk.Label
    body_scrolledwindow: Gtk.ScrolledWindow
    body_textview: Gtk.TextView
    items_list_treeview: Gtk.TreeView
    treeview_selection1: Gtk.TreeSelection
    cancel_button: Gtk.Button
    accept_button: Gtk.Button

class SearchViewBuilder(Builder):
    calendar_popover: Gtk.Popover
    calendar: Gtk.Calendar
    first_day_button: Gtk.Button
    previous_day_button: Gtk.Button
    next_day_button: Gtk.Button
    last_day_button: Gtk.Button
    search_box: Gtk.Box
    search_entry: Gtk.SearchEntry
    search_button: Gtk.Button
    calendar_button: Gtk.MenuButton
    close_button: Gtk.Button
    search_checkbutton: Gtk.CheckButton
    search_filters_box: Gtk.Box
    results_scrolled: Gtk.ScrolledWindow
    results_listbox: Gtk.ListBox
    status_page: Adw.StatusPage
    filter_date_before_popover: Gtk.Popover
    filter_date_before_calendar: Gtk.Calendar
    filter_date_before_reset_button: Gtk.Button
    filter_date_after_popover: Gtk.Popover
    filter_date_after_calendar: Gtk.Calendar
    filter_date_after_reset_button: Gtk.Button
    search_filters_grid: Gtk.Grid
    filter_from_desc_label: Gtk.Label
    filter_from_dropdown: GajimDropDown[ChatDirection | None]
    filter_from_entry: Gtk.Entry
    search_all_from_checkbutton: Gtk.CheckButton
    filter_before_button: Gtk.MenuButton
    filter_after_button: Gtk.MenuButton
    filter_before_label: Gtk.Label
    filter_after_label: Gtk.Label
    header_box: Gtk.Box
    header_name_label: Gtk.Label
    header_date_label: Gtk.Label
    result_row_grid: Gtk.Grid
    row_avatar: Gtk.Image
    row_time_label: Gtk.Label
    row_name_label: Gtk.Label

class ServiceDiscoveryWindowBuilder(Builder):
    service_discovery: Gtk.Box
    banner_agent_icon: Gtk.Image
    banner_agent_header: Gtk.Label
    banner_agent_subheader: Gtk.Label
    address_box: Gtk.Box
    address_comboboxtext: Gtk.ComboBoxText
    browse_button: Gtk.Button
    services_progressbar: Gtk.ProgressBar
    services_scrollwin: Gtk.ScrolledWindow
    services_treeview: Gtk.TreeView
    treeview_selection1: Gtk.TreeSelection
    action_buttonbox: Gtk.Box

class SslErrorDialogBuilder(Builder):
    ssl_error_box: Gtk.Box
    intro_text: Gtk.Label
    ssl_error: Gtk.Label
    add_certificate_checkbutton: Gtk.CheckButton
    view_cert_button: Gtk.Button
    connect_button: Gtk.Button

class StartChatDialogBuilder(Builder):
    stack: Gtk.Stack
    infobar: Gtk.Revealer
    infobar_close_button: Gtk.Button
    box: Gtk.Box
    controls_box: Gtk.Box
    search_entry: Gtk.SearchEntry
    global_search_toggle: Gtk.ToggleButton
    settings_menu: Gtk.MenuButton
    search_error_box: Gtk.Box
    search_error_label: Gtk.Label
    list_stack: Gtk.Stack
    contact_scrolled: Gtk.ScrolledWindow
    no_contacts_placeholder: Gtk.Box
    global_scrolled: Gtk.ScrolledWindow
    global_search_placeholder_stack: Gtk.Stack
    global_search_placeholder_hints: Gtk.Box
    global_search_results_label: Gtk.Label
    spinner: Adw.Spinner
    error_label: Gtk.Label
    error_back_button: Gtk.Button
    info_box: Gtk.Box
    info_back_button: Gtk.Button
    join_box: Gtk.Box
    join_button: Gtk.Button
    account_view: Gtk.TreeView
    icon_pixbuf: Gtk.CellRendererPixbuf
    account_text: Gtk.CellRendererText
    account_back_button: Gtk.Button
    account_select_button: Gtk.Button

class ThemesWindowBuilder(Builder):
    option_popover: Gtk.Popover
    choose_option_listbox: Gtk.ListBox
    placeholder: Gtk.Box
    theme_store: Gtk.ListStore
    theme_grid: Gtk.Grid
    theme_treeview: Gtk.TreeView
    theme_treeview_selection: Gtk.TreeSelection
    theme_name_cell_renderer: Gtk.CellRendererText
    option_listbox: Gtk.ListBox
    add_option_button: Gtk.MenuButton
    add_theme_button: Gtk.Button
    remove_theme_button: Gtk.Button

class VideoPreviewBuilder(Builder):
    video_preview_box: Gtk.Box
    video_source_label: Gtk.Label
    video_preview_placeholder: Gtk.Box

class VoiceMessageRecorderBuilder(Builder):
    popover: Gtk.Popover
    box: Gtk.Box
    error_label: Gtk.Label
    progression_box: Gtk.Box
    time_label: Gtk.Label
    visualization_box: Gtk.Box
    audio_player_box: Gtk.Box
    record_control_box: Gtk.Box
    cancel_button: Gtk.Button
    record_toggle_button: Gtk.Button
    record_toggle_button_image: Gtk.Image
    send_button: Gtk.Button

class WorkspaceDialogBuilder(Builder):
    box: Gtk.Box
    preview: Gtk.Image
    entry: Gtk.Entry
    remove_workspace_button: Gtk.Button
    image_switch: Gtk.Switch
    style_stack: Gtk.Stack
    color_dialog_button: Gtk.ColorDialogButton
    image_box: Gtk.Box
    cancel_button: Gtk.Button
    save_button: Gtk.Button

@overload
def get_builder(
    file_name: Literal["activity_change_timezone.ui"],
    instance: Any = None,
    widgets: list[str] = ...,
) -> ActivityChangeTimezoneBuilder: ...  # noqa
@overload
def get_builder(
    file_name: Literal["activity_default.ui"],
    instance: Any = None,
    widgets: list[str] = ...,
) -> ActivityDefaultBuilder: ...  # noqa
@overload
def get_builder(
    file_name: Literal["activity_gajim_update.ui"],
    instance: Any = None,
    widgets: list[str] = ...,
) -> ActivityGajimUpdateBuilder: ...  # noqa
@overload
def get_builder(
    file_name: Literal["activity_muc_invitation.ui"],
    instance: Any = None,
    widgets: list[str] = ...,
) -> ActivityMucInvitationBuilder: ...  # noqa
@overload
def get_builder(
    file_name: Literal["activity_openpgp_event.ui"],
    instance: Any = None,
    widgets: list[str] = ...,
) -> ActivityOpenpgpEventBuilder: ...  # noqa
@overload
def get_builder(
    file_name: Literal["activity_subscription.ui"],
    instance: Any = None,
    widgets: list[str] = ...,
) -> ActivitySubscriptionBuilder: ...  # noqa
@overload
def get_builder(
    file_name: Literal["advanced_configuration.ui"],
    instance: Any = None,
    widgets: list[str] = ...,
) -> AdvancedConfigurationBuilder: ...  # noqa
@overload
def get_builder(
    file_name: Literal["assistant.ui"], instance: Any = None, widgets: list[str] = ...
) -> AssistantBuilder: ...  # noqa
@overload
def get_builder(
    file_name: Literal["call_window.ui"], instance: Any = None, widgets: list[str] = ...
) -> CallWindowBuilder: ...  # noqa
@overload
def get_builder(
    file_name: Literal["chat_control.ui"],
    instance: Any = None,
    widgets: list[str] = ...,
) -> ChatControlBuilder: ...  # noqa
@overload
def get_builder(
    file_name: Literal["chat_list_row.ui"],
    instance: Any = None,
    widgets: list[str] = ...,
) -> ChatListRowBuilder: ...  # noqa
@overload
def get_builder(
    file_name: Literal["contact_info.ui"],
    instance: Any = None,
    widgets: list[str] = ...,
) -> ContactInfoBuilder: ...  # noqa
@overload
def get_builder(
    file_name: Literal["crypto_trust_manager.ui"],
    instance: Any = None,
    widgets: list[str] = ...,
) -> CryptoTrustManagerBuilder: ...  # noqa
@overload
def get_builder(
    file_name: Literal["db_migration.ui"],
    instance: Any = None,
    widgets: list[str] = ...,
) -> DbMigrationBuilder: ...  # noqa
@overload
def get_builder(
    file_name: Literal["debug_console.ui"],
    instance: Any = None,
    widgets: list[str] = ...,
) -> DebugConsoleBuilder: ...  # noqa
@overload
def get_builder(
    file_name: Literal["exception_dialog.ui"],
    instance: Any = None,
    widgets: list[str] = ...,
) -> ExceptionDialogBuilder: ...  # noqa
@overload
def get_builder(
    file_name: Literal["file_transfer.ui"],
    instance: Any = None,
    widgets: list[str] = ...,
) -> FileTransferBuilder: ...  # noqa
@overload
def get_builder(
    file_name: Literal["file_transfer_jingle.ui"],
    instance: Any = None,
    widgets: list[str] = ...,
) -> FileTransferJingleBuilder: ...  # noqa
@overload
def get_builder(
    file_name: Literal["file_transfer_selector.ui"],
    instance: Any = None,
    widgets: list[str] = ...,
) -> FileTransferSelectorBuilder: ...  # noqa
@overload
def get_builder(
    file_name: Literal["groupchat_affiliation.ui"],
    instance: Any = None,
    widgets: list[str] = ...,
) -> GroupchatAffiliationBuilder: ...  # noqa
@overload
def get_builder(
    file_name: Literal["groupchat_blocks.ui"],
    instance: Any = None,
    widgets: list[str] = ...,
) -> GroupchatBlocksBuilder: ...  # noqa
@overload
def get_builder(
    file_name: Literal["groupchat_config.ui"],
    instance: Any = None,
    widgets: list[str] = ...,
) -> GroupchatConfigBuilder: ...  # noqa
@overload
def get_builder(
    file_name: Literal["groupchat_details.ui"],
    instance: Any = None,
    widgets: list[str] = ...,
) -> GroupchatDetailsBuilder: ...  # noqa
@overload
def get_builder(
    file_name: Literal["groupchat_info_scrolled.ui"],
    instance: Any = None,
    widgets: list[str] = ...,
) -> GroupchatInfoScrolledBuilder: ...  # noqa
@overload
def get_builder(
    file_name: Literal["groupchat_inviter.ui"],
    instance: Any = None,
    widgets: list[str] = ...,
) -> GroupchatInviterBuilder: ...  # noqa
@overload
def get_builder(
    file_name: Literal["groupchat_manage.ui"],
    instance: Any = None,
    widgets: list[str] = ...,
) -> GroupchatManageBuilder: ...  # noqa
@overload
def get_builder(
    file_name: Literal["groupchat_outcast.ui"],
    instance: Any = None,
    widgets: list[str] = ...,
) -> GroupchatOutcastBuilder: ...  # noqa
@overload
def get_builder(
    file_name: Literal["groupchat_roster.ui"],
    instance: Any = None,
    widgets: list[str] = ...,
) -> GroupchatRosterBuilder: ...  # noqa
@overload
def get_builder(
    file_name: Literal["groupchat_roster_tooltip.ui"],
    instance: Any = None,
    widgets: list[str] = ...,
) -> GroupchatRosterTooltipBuilder: ...  # noqa
@overload
def get_builder(
    file_name: Literal["groupchat_state.ui"],
    instance: Any = None,
    widgets: list[str] = ...,
) -> GroupchatStateBuilder: ...  # noqa
@overload
def get_builder(
    file_name: Literal["history_export.ui"],
    instance: Any = None,
    widgets: list[str] = ...,
) -> HistoryExportBuilder: ...  # noqa
@overload
def get_builder(
    file_name: Literal["manage_sounds.ui"],
    instance: Any = None,
    widgets: list[str] = ...,
) -> ManageSoundsBuilder: ...  # noqa
@overload
def get_builder(
    file_name: Literal["message_actions_box.ui"],
    instance: Any = None,
    widgets: list[str] = ...,
) -> MessageActionsBoxBuilder: ...  # noqa
@overload
def get_builder(
    file_name: Literal["password_dialog.ui"],
    instance: Any = None,
    widgets: list[str] = ...,
) -> PasswordDialogBuilder: ...  # noqa
@overload
def get_builder(
    file_name: Literal["pep_config.ui"], instance: Any = None, widgets: list[str] = ...
) -> PepConfigBuilder: ...  # noqa
@overload
def get_builder(
    file_name: Literal["profile.ui"], instance: Any = None, widgets: list[str] = ...
) -> ProfileBuilder: ...  # noqa
@overload
def get_builder(
    file_name: Literal["quit_dialog.ui"], instance: Any = None, widgets: list[str] = ...
) -> QuitDialogBuilder: ...  # noqa
@overload
def get_builder(
    file_name: Literal["roster_item_exchange.ui"],
    instance: Any = None,
    widgets: list[str] = ...,
) -> RosterItemExchangeBuilder: ...  # noqa
@overload
def get_builder(
    file_name: Literal["search_view.ui"], instance: Any = None, widgets: list[str] = ...
) -> SearchViewBuilder: ...  # noqa
@overload
def get_builder(
    file_name: Literal["service_discovery_window.ui"],
    instance: Any = None,
    widgets: list[str] = ...,
) -> ServiceDiscoveryWindowBuilder: ...  # noqa
@overload
def get_builder(
    file_name: Literal["ssl_error_dialog.ui"],
    instance: Any = None,
    widgets: list[str] = ...,
) -> SslErrorDialogBuilder: ...  # noqa
@overload
def get_builder(
    file_name: Literal["start_chat_dialog.ui"],
    instance: Any = None,
    widgets: list[str] = ...,
) -> StartChatDialogBuilder: ...  # noqa
@overload
def get_builder(
    file_name: Literal["themes_window.ui"],
    instance: Any = None,
    widgets: list[str] = ...,
) -> ThemesWindowBuilder: ...  # noqa
@overload
def get_builder(
    file_name: Literal["video_preview.ui"],
    instance: Any = None,
    widgets: list[str] = ...,
) -> VideoPreviewBuilder: ...  # noqa
@overload
def get_builder(
    file_name: Literal["voice_message_recorder.ui"],
    instance: Any = None,
    widgets: list[str] = ...,
) -> VoiceMessageRecorderBuilder: ...  # noqa
@overload
def get_builder(
    file_name: Literal["workspace_dialog.ui"],
    instance: Any = None,
    widgets: list[str] = ...,
) -> WorkspaceDialogBuilder: ...  # noqa
def get_builder(
    file_name: str, instance: Any = None, widgets: list[str] = ...
) -> Builder: ...



# This file is part of Gajim.
#
# SPDX-License-Identifier: GPL-3.0-only

# XEP-0145: Annotations


from nbxmpp.errors import MalformedStanzaError
from nbxmpp.errors import StanzaError
from nbxmpp.protocol import JID
from nbxmpp.structs import AnnotationNote
from nbxmpp.task import Task

from gajim.common.modules.base import BaseModule
from gajim.common.types import ConnectionT


class Annotations(BaseModule):
    _nbxmpp_extends = "Annotations"
    _nbxmpp_methods = [
        "request_annotations",
        "set_annotations",
    ]

    def __init__(self, con: ConnectionT) -> None:
        BaseModule.__init__(self, con)

        self._annotations: dict[JID | str, AnnotationNote] = {}

    def request_annotations(self) -> None:
        self._nbxmpp("Annotations").request_annotations(
            callback=self._annotations_received
        )

    def _annotations_received(self, task: Task) -> None:
        try:
            annotations = task.finish()
        except (StanzaError, MalformedStanzaError) as error:
            self._log.warning(error)
            self._annotations = {}
            return

        for note in annotations:
            self._annotations[note.jid] = note

    def get_note(self, jid: JID | str) -> AnnotationNote | None:
        return self._annotations.get(jid)

    def set_note(self, note: AnnotationNote) -> None:
        self._annotations[note.jid] = note
        self.set_annotations(self._annotations.values())



# This file is part of Gajim.
#
# SPDX-License-Identifier: GPL-3.0-only

from __future__ import annotations

from typing import cast
from typing import NamedTuple

import logging
from enum import IntEnum

from gi.repository import Gtk
from nbxmpp.errors import StanzaError
from nbxmpp.structs import AffiliationResult
from nbxmpp.task import Task

from gajim.common import app
from gajim.common.client import Client
from gajim.common.i18n import _
from gajim.common.modules.contacts import GroupchatContact

from gajim.gtk.alert import InformationAlertDialog
from gajim.gtk.apply_button_box import ApplyButtonBox
from gajim.gtk.builder import get_builder
from gajim.gtk.util.classes import SignalManager

log = logging.getLogger("gajim.gtk.groupchat_outcasts")


class OutcastRow(NamedTuple):
    jid: str
    reason: str | None


class Column(IntEnum):
    JID = 0
    REASON = 1


class GroupchatOutcasts(Gtk.Box, SignalManager):
    def __init__(self, client: Client, contact: GroupchatContact) -> None:
        Gtk.Box.__init__(self, orientation=Gtk.Orientation.VERTICAL)
        SignalManager.__init__(self)

        self._client = client
        self._contact = contact
        self._current_rows: set[OutcastRow] = set()

        self_contact = contact.get_self()
        assert self_contact is not None
        self._own_affiliation = self_contact.affiliation.value

        self._ui = get_builder("groupchat_outcast.ui")

        self._connect(self._ui.outcast_selection, "changed", self._on_selection_changed)
        self._connect(self._ui.address_renderer, "edited", self._on_jid_edited)
        self._connect(self._ui.reason_renderer, "edited", self._on_reason_edited)
        self._connect(self._ui.add_button, "clicked", self._on_add)
        self._connect(self._ui.remove_button, "clicked", self._on_remove)

        self._treeview = self._ui.outcast_treeview
        self._store = self._ui.outcast_store

        self.append(self._ui.main_box)

        self._apply_button = ApplyButtonBox(_("Apply"), on_clicked=self._on_apply)
        self._apply_button.set_hexpand(True)
        self._apply_button.set_halign(Gtk.Align.END)

        self._ui.button_box.append(self._apply_button)

        if self._own_affiliation in ("admin", "owner"):
            self._ui.add_button.set_sensitive(True)
            self._ui.add_button.set_tooltip_text("")

        self._client.get_module("MUC").get_affiliation(
            self._contact.jid, "outcast", callback=self._on_outcasts_received
        )

    def do_unroot(self) -> None:
        self._disconnect_all()
        del self._treeview
        del self._store
        del self._apply_button
        Gtk.Box.do_unroot(self)
        app.check_finalize(self)

    def _on_apply(self, _button: Gtk.Button) -> None:
        self._begin_progress()
        self._set_outcasts()

    def _begin_progress(self) -> None:
        self._ui.outcast_scrolled.set_sensitive(False)
        self._ui.add_remove_button_box.set_sensitive(False)

    def _end_progress(self) -> None:
        self._ui.outcast_scrolled.set_sensitive(True)
        self._ui.add_remove_button_box.set_sensitive(True)

    def _on_add(self, _button: Gtk.Button) -> None:
        iter_ = self._store.append([None, None, True])

        path = self._store.get_path(iter_)
        self._treeview.scroll_to_cell(path, None, False, 0, 0)
        self._treeview.get_selection().unselect_all()
        self._treeview.get_selection().select_path(path)

        self._update_apply_button_state()

    def _on_remove(self, _button: Gtk.Button) -> None:
        selection = self._treeview.get_selection()
        assert selection is not None
        rows = selection.get_selected_rows()
        if rows is None:
            return

        _model, paths = rows

        references: list[Gtk.TreeRowReference] = []
        for path in paths:
            references.append(Gtk.TreeRowReference(self._store, path))

        for ref in references:
            path = ref.get_path()
            assert path is not None
            iter_ = self._store.get_iter(path)
            self._store.remove(iter_)

        self._update_apply_button_state()

    def _on_jid_edited(
        self, _renderer: Gtk.CellRendererText, path: str, new_text: str
    ) -> None:
        old_text = self._store[path][Column.JID]
        if new_text == old_text:
            return

        if self._jid_exists(new_text):
            self._raise_error()
            return

        self._store[path][Column.JID] = new_text
        self._update_apply_button_state()

    def _on_reason_edited(
        self, _renderer: Gtk.CellRendererText, path: str, new_text: str
    ) -> None:
        self._store[path][Column.REASON] = new_text or None
        self._update_apply_button_state()

    def _on_selection_changed(self, tree_selection: Gtk.TreeSelection) -> None:
        sensitive = bool(tree_selection.count_selected_rows())
        self._set_remove_button_state(sensitive)

    def _jid_exists(self, jid: str) -> bool:
        return any(row[Column.JID] == jid for row in self._store)

    def _set_remove_button_state(self, sensitive: bool) -> None:
        value = self._own_affiliation in ("admin", "owner")
        self._ui.remove_button.set_sensitive(sensitive and value)

    def _update_apply_button_state(self):
        if self._own_affiliation not in ("admin", "owner"):
            return
        new_rows = self._get_new_rows()
        self._apply_button.set_button_state(new_rows != self._current_rows)

    def _allowed_to_edit(self) -> bool:
        return self._own_affiliation in ("owner", "admin")

    def _get_new_rows(self) -> set[OutcastRow]:
        rows: set[OutcastRow] = set()

        for row in self._store:
            if not row[Column.JID]:
                continue

            rows.add(OutcastRow(jid=row[Column.JID], reason=row[Column.REASON]))
        return rows

    def _get_diff(self) -> tuple[list[OutcastRow], list[OutcastRow]]:
        new_rows = self._get_new_rows()

        before = {row.jid for row in self._current_rows}
        after = {row.jid for row in new_rows}

        removed = before - after
        removed_rows = [row for row in self._current_rows if row.jid in removed]

        added = after - before
        added_rows = [row for row in new_rows if row.jid in added]

        same = after - removed - added
        same_rows = {row for row in new_rows if row.jid in same}
        modified_rows = list(same_rows - self._current_rows)

        return removed_rows, added_rows + modified_rows

    def _set_outcasts(self) -> None:
        removed_rows, other_rows = self._get_diff()

        outcasts = {}
        for row in other_rows:
            outcasts[row.jid] = {"affiliation": "outcast"}
            if row.reason:
                outcasts[row.jid]["reason"] = row.reason

        for row in removed_rows:
            outcasts[row.jid] = {"affiliation": "none"}

        self._client.get_module("MUC").set_affiliation(
            self._contact.jid, outcasts, callback=self._on_affiliation_finished
        )

    def _on_affiliation_finished(self, task: Task) -> None:
        try:
            task.finish()
        except StanzaError as error:
            log.info("Error while setting outcasts: %s", error)
            self._end_progress()
            self._apply_button.set_error(str(error))
            return

        self._current_rows = self._get_new_rows()
        self._end_progress()
        self._apply_button.set_success()

    def _on_outcasts_received(self, task: Task) -> None:
        try:
            result = cast(AffiliationResult, task.finish())
        except StanzaError as error:
            log.info("Error while requesting outcasts: %s", error.condition)
            return

        editable = self._allowed_to_edit()
        for jid, attrs in result.users.items():
            reason = attrs.get("reason")

            jid = str(jid)
            self._store.append([jid, reason, editable])

            self._current_rows.add(OutcastRow(jid=jid, reason=reason))

    @staticmethod
    def _raise_error() -> None:
        InformationAlertDialog(
            _("Error"), _("An entry with this XMPP Address already exists")
        )



# This file is part of Gajim.
#
# SPDX-License-Identifier: GPL-3.0-only

from typing import Any

import logging

from gi.repository import Gtk
from gi.repository import GtkSource

from gajim.common import app
from gajim.common import ged
from gajim.common.i18n import _
from gajim.common.styling import PreBlock

from gajim.gtk.util.classes import SignalManager
from gajim.gtk.util.styling import get_source_view_style_scheme

log = logging.getLogger("gajim.gtk.conversation.code_widget")

LANG_MAP: dict[str | None, str | None] = {"python": "python3"}


class CodeWidget(Gtk.Box, SignalManager):
    def __init__(self, account: str) -> None:
        Gtk.Box.__init__(self, orientation=Gtk.Orientation.VERTICAL, vexpand=True)
        SignalManager.__init__(self)

        self.add_css_class("code-widget")

        self._account = account

        header = Gtk.Box()
        header.set_spacing(6)
        header.add_css_class("code-widget-header")
        self._lang_label = Gtk.Label()
        header.append(self._lang_label)

        copy_button = Gtk.Button.new_from_icon_name("lucide-copy-symbolic")
        copy_button.set_tooltip_text(_("Copy code snippet"))
        copy_button.set_halign(Gtk.Align.END)
        copy_button.set_hexpand(True)
        self._connect(copy_button, "clicked", self._on_copy)
        header.append(copy_button)
        self.append(header)

        self._textview = CodeTextview()
        self._scrolled = Gtk.ScrolledWindow()
        self._scrolled.set_policy(Gtk.PolicyType.AUTOMATIC, Gtk.PolicyType.NEVER)
        self._scrolled.set_hexpand(True)
        self._scrolled.set_vexpand(True)
        self._scrolled.set_propagate_natural_width(True)
        self._scrolled.set_propagate_natural_height(True)
        self._scrolled.set_max_content_height(400)
        self._scrolled.set_child(self._textview)

        self.append(self._scrolled)

    def do_unroot(self) -> None:
        self._disconnect_all()

        Gtk.Box.do_unroot(self)
        app.check_finalize(self)

    def _on_copy(self, _button: Gtk.Button) -> None:
        self.get_clipboard().set(self._textview.get_code())

    def add_content(self, block: PreBlock):
        code, lang = self._prepare_code(block.text)
        if lang is None:
            self._lang_label.set_text(_("Code snippet"))
        else:
            lang_name = self._textview.set_language(lang)
            self._lang_label.set_text(lang_name)

        self._textview.print_code(code)

    @staticmethod
    def _prepare_code(text: str) -> tuple[str, str | None]:
        text = text.strip()
        code_start = text.partition("\n")[0]
        lang = None
        if len(code_start) > 3:
            lang = code_start[3:]

        code = text.partition("\n")[2][:-4]
        return code, LANG_MAP.get(lang, lang)


class CodeTextview(GtkSource.View):
    def __init__(self) -> None:
        GtkSource.View.__init__(self)
        self.set_editable(False)
        self.set_cursor_visible(False)
        self.set_top_margin(2)
        self.set_bottom_margin(2)
        self.set_monospace(True)
        self.get_buffer().set_highlight_matching_brackets(False)

        self._source_manager = GtkSource.LanguageManager.get_default()

        app.ged.register_event_handler(
            "style-changed", ged.GUI1, self._on_style_changed
        )

        style_scheme = get_source_view_style_scheme()
        if style_scheme is not None:
            self.get_buffer().set_style_scheme(style_scheme)

    def do_unroot(self) -> None:
        app.ged.remove_event_handler("style-changed", ged.GUI1, self._on_style_changed)

        GtkSource.View.do_unroot(self)
        app.check_finalize(self)

    def _on_style_changed(self, *args: Any) -> None:
        style_scheme = get_source_view_style_scheme()
        if style_scheme is not None:
            self.get_buffer().set_style_scheme(style_scheme)

    def set_language(self, language_string: str) -> str:
        lang = self._source_manager.get_language(language_string)
        if lang is None:
            return _("Unknown language")

        log.debug("Code snippet lang: %s", lang.get_name())
        self.get_buffer().set_language(lang)
        return lang.get_name()

    def get_code(self) -> str:
        buffer_ = self.get_buffer()
        start, end = buffer_.get_bounds()
        return buffer_.get_text(start, end, False)

    def print_code(self, code: str) -> None:
        buffer_ = self.get_buffer()
        buffer_.insert(buffer_.get_start_iter(), code)

        if buffer_.get_line_count() == 1:
            # Add some space at the bottom so that scrollbar
            # does not overlap content, which makes text hard to select
            self.set_pixels_below_lines(10)



# This file is part of Gajim.
#
# SPDX-License-Identifier: GPL-3.0-only

from __future__ import annotations

from typing import Literal

from enum import IntEnum


class MessageState(IntEnum):
    PENDING = 1
    ACKNOWLEDGED = 2


class MessageType(IntEnum):
    CHAT = 1
    GROUPCHAT = 2
    PM = 3

    @classmethod
    def from_str(cls, string: str) -> MessageType:
        return cls[string.upper()]

    def to_str(self) -> Literal["chat", "groupchat", "pm"]:
        return self.name.lower()  # type: ignore


class ChatDirection(IntEnum):
    INCOMING = 1
    OUTGOING = 2



# This file is part of Gajim.
#
# SPDX-License-Identifier: GPL-3.0-only

from datetime import datetime

from gi.repository import Gtk

from gajim.common import app

from gajim.gtk.conversation.rows.base import BaseRow


class DateRow(BaseRow):
    def __init__(self, account: str, timestamp: datetime) -> None:
        BaseRow.__init__(self, account, widget="label")

        self.set_selectable(False)
        self.set_activatable(False)

        self.type = "date"
        self.timestamp = timestamp
        self.add_css_class("conversation-date-row")

        format_string = app.settings.get("date_format")
        self.label.set_text(timestamp.strftime("%a " + format_string))
        self.label.set_halign(Gtk.Align.CENTER)
        self.label.set_hexpand(True)
        self.label.add_css_class("conversation-meta")
        self.grid.attach(self.label, 0, 0, 1, 1)

    def do_unroot(self) -> None:
        BaseRow.do_unroot(self)



# This file is part of Gajim.
#
# SPDX-License-Identifier: GPL-3.0-only

from __future__ import annotations

from typing import Any

from gi.repository import Adw
from gi.repository import GLib
from gi.repository import GObject

from gajim.common import app
from gajim.common import events
from gajim.common import ged
from gajim.common.client import Client
from gajim.common.const import SimpleClientState
from gajim.common.ged import EventHelper
from gajim.common.i18n import _
from gajim.common.util.text import to_one_line


class StatusMessageSelectorRow(Adw.EntryRow, EventHelper):
    __gtype_name__ = "StatusMessageSelectorRow"

    def __init__(self) -> None:
        Adw.EntryRow.__init__(self)
        EventHelper.__init__(self)

        # These need to be set after init (Adw issue)
        self.set_title(_("Status Message"))
        self.set_show_apply_button(True)
        self.set_enable_emoji_completion(True)

        self.connect("apply", self._on_apply_clicked)

        self._account: str | None = None

        self.register_events(
            [
                ("our-show", ged.GUI1, self._on_our_show),
            ]
        )

    @GObject.Property(type=str)
    def account(self) -> str | None:  # type: ignore
        return self._account

    @account.setter
    def account(self, account: str | None) -> None:
        if self._account == account:
            return

        self._disconnect_signals(self._account)
        self._connect_signals(account)

        self.set_sensitive(account is not None)

        self._account = account
        self._update_status_message()

    def set_account(self, account: str | None) -> None:
        self.account = account

    def _connect_signals(self, account: str | None) -> None:
        if account is None:
            return
        client = app.get_client(account)
        client.connect_signal("state-changed", self._on_client_state_changed)

    def _disconnect_signals(self, account: str | None) -> None:
        if account is None:
            return
        client = app.get_client(account)
        client.disconnect_all_from_obj(self)

    def _on_our_show(self, event: events.ShowChanged) -> None:
        if event.account != self._account:
            return
        self._update_status_message()

    def _on_client_state_changed(
        self, client: Client, _signal_name: str, _state: SimpleClientState
    ) -> None:
        self._update_status_message()

    def _on_apply_clicked(self, *args: Any) -> None:
        self.set_show_apply_button(False)
        message = self.get_text()
        message = to_one_line(message)

        assert self._account is not None

        for client in app.get_clients():
            sync = app.settings.get_account_setting(
                client.account, "sync_with_global_status"
            )
            if client.account != self._account and not sync:
                continue
            client.change_status(client.status, message)

        self.set_show_apply_button(False)
        GLib.timeout_add_seconds(3, self._show_apply_button)

    def _show_apply_button(self) -> None:
        self.set_show_apply_button(True)

    def _update_status_message(self) -> None:
        if self._account is None:
            self.set_text("")
            return

        client = app.get_client(self._account)
        self.set_text(client.status_message)



# Copyright (C) 2010-2011 Denis Fomin <fominde AT gmail.com>
#
# This file is part of Gajim.
#
# SPDX-License-Identifier: GPL-3.0-only

import gettext
import locale
from pathlib import Path

from gajim.common import configpaths

DOMAIN = "gajim_plugins"
try:
    plugin_user_dir = configpaths.get("PLUGINS_USER")
except KeyError:
    # This allows to import the module for tests
    print("No plugin translation path available")
    plugin_user_dir = Path.cwd()


plugins_locale_dir = plugin_user_dir / "locale"

try:
    t = gettext.translation(DOMAIN, plugins_locale_dir)
    _ = t.gettext
except OSError:
    _ = gettext.gettext

if hasattr(locale, "bindtextdomain"):
    locale.bindtextdomain(DOMAIN, plugins_locale_dir)



# This file is part of Gajim.
#
# SPDX-License-Identifier: GPL-3.0-only

from __future__ import annotations

from typing import cast

import locale
import logging

from gi.repository import GObject
from gi.repository import Gtk
from nbxmpp import JID
from nbxmpp.structs import DiscoInfo

from gajim.common import app
from gajim.common import types
from gajim.common.i18n import _
from gajim.common.modules.contacts import BareContact
from gajim.common.modules.contacts import ResourceContact

from gajim.gtk.util.classes import SignalManager
from gajim.gtk.util.misc import clear_listbox

log = logging.getLogger("gajim.gtk.resource_selector")


class ResourceSelector(Gtk.ScrolledWindow, SignalManager):
    __gsignals__ = {
        "selection-changed": (GObject.SignalFlags.RUN_LAST, None, (bool,)),
    }

    def __init__(
        self, contact: BareContact, constraints: list[str] | None = None
    ) -> None:
        Gtk.ScrolledWindow.__init__(self, height_request=200)
        SignalManager.__init__(self)

        self.add_css_class("resource-selector")

        self._listbox = Gtk.ListBox()
        self._listbox.set_sort_func(self._sort_func)
        self._connect(self._listbox, "row-selected", self._on_row_selected)
        self.set_child(self._listbox)

        self._contact = contact
        self._contact.connect("presence-update", self._on_update)
        self._contact.connect("caps-update", self._on_update)

        # Constraints include nbxmpp Namespaces a resource has to support
        self._constraints = constraints or []

        self._set_placeholder()
        self._add_entries()

    def do_unroot(self) -> None:
        self._disconnect_all()
        self._contact.disconnect_all_from_obj(self)

        self._listbox.set_sort_func(None)
        Gtk.ScrolledWindow.do_unroot(self)
        app.check_finalize(self)

    @staticmethod
    def _sort_func(row1: ResourceRow, row2: ResourceRow) -> int:
        return locale.strcoll(row1.device_text.lower(), row2.device_text.lower())

    def _on_row_selected(self, _listbox: Gtk.ListBox, row: ResourceRow | None) -> None:
        state = bool(row is not None)
        self.emit("selection-changed", state)

    def _set_placeholder(self) -> None:
        image = Gtk.Image.new_from_icon_name("lucide-circle-alert-symbolic")
        label = Gtk.Label(label=_("No devices online"))
        box = Gtk.Box(
            orientation=Gtk.Orientation.VERTICAL, spacing=6, valign=Gtk.Align.CENTER
        )
        box.add_css_class("dimmed")
        box.append(image)
        box.append(label)
        self._listbox.set_placeholder(box)

    def _add_entries(self) -> None:
        for resource in self._contact.iter_resources():
            self._listbox.append(ResourceRow(resource, self._constraints))

    def _on_update(self, _contact: types.ResourceContact, _signal_name: str) -> None:
        clear_listbox(self._listbox)
        self._add_entries()

    def get_jid(self) -> JID:
        resource_row = cast(ResourceRow, self._listbox.get_selected_row())
        return resource_row.jid


class ResourceRow(Gtk.ListBoxRow):
    def __init__(
        self, resource_contact: ResourceContact, constraints: list[str]
    ) -> None:
        Gtk.ListBoxRow.__init__(self)

        self.jid = resource_contact.jid

        icon_name = "lucide-laptop-symbolic"
        tooltip_text = _("Computer")
        self.device_text = resource_contact.jid.resource or ""

        disco_info = app.storage.cache.get_last_disco_info(resource_contact.jid)
        if disco_info is not None:
            name, type_ = self._get_client_identity(disco_info)
            if name is not None:
                self.device_text = f"{name} ({resource_contact.jid.resource})"
            if type_ is not None:
                if type_ == "phone":
                    icon_name = "lucide-smartphone-symbolic"
                    tooltip_text = _("Phone")

        image = Gtk.Image(icon_name=icon_name, tooltip_text=tooltip_text, pixel_size=32)

        name_label = Gtk.Label()
        name_label.set_text(self.device_text)

        box = Gtk.Box(spacing=12)
        box.append(image)
        box.append(name_label)

        for constraint in constraints:
            if not resource_contact.supports(constraint):
                self.set_sensitive(False)
                self.set_tooltip_text(_("This device is not compatible."))

        self.set_child(box)

    @staticmethod
    def _get_client_identity(disco_info: DiscoInfo) -> tuple[str | None, str | None]:
        for identity in disco_info.identities:
            if identity.category == "client":
                return identity.name, identity.type
        return None, None



# Copyright (C) 2003-2014 Yann Leboulanger <asterix AT lagaule.org>
# Copyright (C) 2005-2006 Dimitur Kirov <dkirov AT gmail.com>
# Copyright (C) 2005-2007 Nikos Kouremenos <kourem AT gmail.com>
# Copyright (C) 2006 Travis Shirk <travis AT pobox.com>
#
# This file is part of Gajim.
#
# SPDX-License-Identifier: GPL-3.0-only

# from __future__ import annotations

# import logging
# import os
# import time
# import uuid
# from datetime import datetime
# from datetime import timezone
# from enum import IntEnum
# from enum import unique
# from functools import partial
# from pathlib import Path

# from gi.repository import Gdk
# from gi.repository import GLib
# from gi.repository import Gtk
# from gi.repository import Pango
# from nbxmpp.protocol import JID

# from gajim.common import app
# from gajim.common import ged
# from gajim.common import types
# from gajim.common.events import FileRequestSent
# from gajim.common.events import Notification
# from gajim.common.file_props import FileProp
# from gajim.common.file_props import FilesProp
# from gajim.common.helpers import file_is_locked
# from gajim.common.i18n import _
# from gajim.common.modules.bytestream import is_transfer_active
# from gajim.common.modules.bytestream import is_transfer_paused
# from gajim.common.modules.bytestream import is_transfer_stopped
# from gajim.common.modules.contacts import BareContact
# from gajim.common.storage.archive import models as mod
# from gajim.common.storage.archive.const import ChatDirection
# from gajim.common.storage.archive.const import MessageState
# from gajim.common.storage.archive.const import MessageType
# from gajim.common.util.datetime import utc_now
# from gajim.common.util.text import get_random_string
# from gajim.gtk.util.misc import open_file

# from gajim.gtk.builder import get_builder
# from gajim.gtk.alert import ConfirmationAlertDialog
# from gajim.gtk.alert import DialogButton
# from gajim.gtk.alert import InformationAlertDialog
# from gajim.gtk.filechoosers import FileSaveDialog
# from gajim.common.util.text import format_eta

# log = logging.getLogger('gajim.gtk.filetransfer')


# @unique
# class Column(IntEnum):
#     IMAGE = 0
#     LABELS = 1
#     FILE = 2
#     TIME = 3
#     PROGRESS = 4
#     PERCENT = 5
#     PULSE = 6
#     SID = 7

# TODO Jingle FT
# Jingle file transfers have been disabled
# class FileTransfersWindow:
#     def __init__(self):
#         self.files_props = {'r': {}, 's': {}}
#         self.height_diff = 0

#         self._last_progress_update = 0

#         self._ui = get_builder('filetransfers.ui', self)
#         self.window = self._ui.file_transfers_window
#         self._ui.notify_ft_complete.set_active(
#             app.settings.get('notify_on_file_complete'))
#         self.model = Gtk.ListStore(str, str, str, str, str, int, int, str)
#         self._ui.transfers_list.set_model(self.model)
#         col = Gtk.TreeViewColumn()

#         render_pixbuf = Gtk.CellRendererPixbuf()

#         col.pack_start(render_pixbuf, True)
#         render_pixbuf.set_property('xpad', 6)
#         render_pixbuf.set_property('ypad', 6)
#         render_pixbuf.set_property('yalign', 0.5)
#         col.add_attribute(render_pixbuf, 'icon_name', 0)
#         self._ui.transfers_list.append_column(col)

#         col = Gtk.TreeViewColumn(_('File'))
#         renderer = Gtk.CellRendererText()
#         col.pack_start(renderer, False)
#         col.add_attribute(renderer, 'markup', Column.LABELS)
#         renderer.set_property('xalign', 0.0)
#         renderer.set_property('yalign', 0.0)
#         renderer = Gtk.CellRendererText()
#         col.pack_start(renderer, True)
#         col.add_attribute(renderer, 'markup', Column.FILE)
#         renderer.set_property('xalign', 0.0)
#         renderer.set_property('yalign', 0.0)
#         renderer.set_property('ellipsize', Pango.EllipsizeMode.END)
#         col.set_resizable(True)
#         col.set_min_width(160)
#         col.set_expand(True)
#         self._ui.transfers_list.append_column(col)

#         col = Gtk.TreeViewColumn(_('Time'))
#         renderer = Gtk.CellRendererText()
#         col.pack_start(renderer, False)
#         col.add_attribute(renderer, 'markup', Column.TIME)
#         renderer.set_property('yalign', 0.5)
#         renderer.set_property('xalign', 0.5)
#         renderer = Gtk.CellRendererText()
#         renderer.set_property('ellipsize', Pango.EllipsizeMode.END)
#         renderer.set_property('xalign', 0.5)
#         col.set_resizable(True)
#         col.set_min_width(70)
#         self._ui.transfers_list.append_column(col)

#         col = Gtk.TreeViewColumn(_('Progress'))
#         renderer = Gtk.CellRendererProgress()
#         renderer.set_property('yalign', 0.5)
#         renderer.set_property('xalign', 0.5)
#         col.pack_start(renderer, False)
#         col.add_attribute(renderer, 'text', Column.PROGRESS)
#         col.add_attribute(renderer, 'value', Column.PERCENT)
#         col.add_attribute(renderer, 'pulse', Column.PULSE)
#         col.set_resizable(False)
#         col.set_fixed_width(150)
#         self._ui.transfers_list.append_column(col)

#         self.icons = {
#             'upload': 'lucide-chevron-up-symbolic',
#             'download': 'lucide-chevron-down-symbolic',
#             'stop': 'lucide-circle-x-symbolic',
#             'waiting': 'lucide-refresh-cw-symbolic',
#             'pause': 'media-playback-pause-symbolic',
#             'continue': 'media-playback-start-symbolic',
#             'ok': 'lucide-check-symbolic',
#             'computing': 'system-run-symbolic',
#             'hash_error': 'network-error-symbolic',
#         }

#         if app.settings.get('use_kib_mib'):
#             self.units = GLib.FormatSizeFlags.IEC_UNITS
#         else:
#             self.units = GLib.FormatSizeFlags.DEFAULT

#         self._ui.transfers_list.get_selection().set_mode(
#             Gtk.SelectionMode.SINGLE)
#         self._ui.transfers_list.get_selection().connect(
#             'changed', self._selection_changed)

#         # Tooltip
#         self._ui.transfers_list.connect('query-tooltip', self._query_tooltip)
#         self._ui.transfers_list.set_has_tooltip(True)
#         self.tooltip = FileTransfersTooltip()

#         self._ui.connect_signals(self)

#         app.ged.register_event_handler(
#             'file-completed', ged.GUI1, self._file_completed)
#         app.ged.register_event_handler(
#             'file-hash-error', ged.GUI1, self._file_hash_error)
#         app.ged.register_event_handler(
#             'file-send-error', ged.GUI1, self._file_send_error)
#         app.ged.register_event_handler(
#             'file-request-error', ged.GUI1, self._file_request_error)
#         app.ged.register_event_handler(
#             'file-request-received', ged.GUI1, self._file_request_received)
#         app.ged.register_event_handler(
#             'file-progress', ged.GUI1, self._file_progress)
#         app.ged.register_event_handler(
#             'file-error', ged.GUI1, self._file_error)
#         app.ged.register_event_handler(
#             'jingle-ft-cancelled-received',
#             ged.GUI1,
#             self._file_transfer_cancelled)

#     def _file_completed(self, event):
#         self.set_status(event.file_props, 'ok')
#         self.set_progress(
#             event.file_props.type_,
#             event.file_props.sid,
#             event.file_props.received_len)

#         if not app.settings.get('notify_on_file_complete'):
#             return

#         app.ged.raise_event(
#             Notification(account=event.account,
#                          jid=event.jid,
#                          type='file-transfer',
#                          sub_type='file-completed',
#                          title=_('File Transfer Completed'),
#                          text=_('File: %s') % event.file_props.name))

#     def _file_error(self, event):
#         self.set_status(event.file_props, 'stop')

#     def _file_hash_error(self, event):
#         self.set_status(event.file_props, 'hash_error')
#         if self.window.get_property('visible'):
#             self.show_hash_error(event.jid, event.file_props, event.account)

#     def _file_send_error(self, event):
#         self.set_status(event.file_props, 'stop')

#         if app.window.is_chat_active(event.account, event.jid):
#             return

#         app.ged.raise_event(
#             Notification(account=event.account,
#                          jid=event.jid,
#                          type='file-transfer',
#                          sub_type='file-send-error',
#                          title=_('File Transfer Failed'),
#                          text=_('File: %s') % event.file_props.name))

#     def _file_request_error(self, event):
#         self.set_status(event.file_props, 'stop')
#         errno = event.file_props.error
#         account = event.conn.name

#         if self.window.get_property('visible'):
#             if errno in (-4, -5):
#                 self.show_stopped(event.jid, event.file_props, event.error_msg)
#             return

#         if errno in (-4, -5):
#             sub_type = 'file-error'
#         else:
#             sub_type = 'file-request-error'

#         if app.window.is_chat_active(account, event.jid):
#             return

#         text = _('File: %s') % event.file_props.name
#         app.ged.raise_event(
#             Notification(account=account,
#                          jid=event.jid,
#                          type='file-transfer',
#                          sub_type=sub_type,
#                          title=_('File Transfer Failed'),
#                          text=text))

#     def _file_request_received(self, event):
#         client = app.get_client(event.account)
#         contact = client.get_module('Contacts').get_contact(event.jid)

#         if event.file_props.session_type == 'jingle':
#             jingle = event.stanza.getTag('jingle')
#             content = jingle.getTag('content')
#             description = content.getTag('description')
#             request = description.getTag('request')
#             if request:
#                 # If we get a request instead
#                 self.add_transfer(event.account, contact, event.file_props)
#                 return

#         if app.window.is_chat_active(event.account, event.jid):
#             return

#         text = _('%s wants to send you a file') % contact.name
#         app.ged.raise_event(
#             Notification(account=event.account,
#                          jid=event.jid,
#                          type='file-transfer',
#                          sub_type='file-request-received',
#                          title=_('File Offered'),
#                          text=text))

#     def _file_transfer_cancelled(self, event):
#         file_props = None
#         # get the file_props of our session
#         file_props = FilesProp.getFileProp(event.conn.name, event.sid)
#         if not file_props:
#             return

#         self.set_status(file_props, 'stop')
#         file_props.error = -4  # is it the right error code?
#         if self.window.get_property('visible'):
#             self.show_stopped(event.jid,
#                               file_props,
#                               'Peer cancelled the transfer')

#     def _query_tooltip(self, widget, x_pos, y_pos, keyboard_mode, tooltip):
#         try:
#             x_pos, y_pos = widget.convert_widget_to_bin_window_coords(
#                 x_pos, y_pos)
#             row = widget.get_path_at_pos(x_pos, y_pos)[0]
#         except TypeError:
#             self.tooltip.clear_tooltip()
#             return False
#         if not row:
#             self.tooltip.clear_tooltip()
#             return False

#         iter_ = None
#         try:
#             model = widget.get_model()
#             iter_ = model.get_iter(row)
#         except Exception:
#             self.tooltip.clear_tooltip()
#             return False

#         sid = self.model[iter_][Column.SID]
#         file_props = FilesProp.getFilePropByType(sid[0], sid[1:])

#         value, widget = self.tooltip.get_tooltip(file_props, sid)
#         tooltip.set_custom(widget)
#         return value

#     def find_transfer_by_jid(self, account, jid):
#         '''
#         Find all transfers with peer 'jid' that belong to 'account'
#         '''
#         active_transfers = [[], []]  # ['senders', 'receivers']
#         allfp = FilesProp.getAllFileProp()
#         for file_props in allfp:
#             if file_props.type_ == 's' and file_props.tt_account == account:
#                 # 'account' is the sender
#                 receiver_jid = file_props.receiver.split('/')[0]
#                 if jid == receiver_jid and not is_transfer_stopped(file_props):
#                     active_transfers[0].append(file_props)
#             elif file_props.type_ == 'r' and file_props.tt_account == account:
#                 # 'account' is the recipient
#                 sender_jid = file_props.sender.split('/')[0]
#                 if jid == sender_jid and not is_transfer_stopped(file_props):
#                     active_transfers[1].append(file_props)
#             else:
#                 raise Exception('file_props has no type')
#         return active_transfers

#     def show_stopped(self, jid, file_props, error_msg=''):
#         if file_props.type_ == 'r':
#             file_name = os.path.basename(file_props.file_name)
#         else:
#             file_name = file_props.name
#         sectext = '\t' + _('Filename: %s') % GLib.markup_escape_text(file_name)
#         sectext += '\n\t' + _('Recipient: %s') % jid
#         if error_msg:
#             sectext += '\n\t' + _('Error message: %s') % error_msg
#         InformationAlertDialog(_('File Transfer Stopped'), sectext)
#         self._ui.transfers_list.get_selection().unselect_all()

#     def show_hash_error(self,
#                         jid: JID,
#                         file_props: FileProp,
#                         account: str
#                         ) -> None:
#         def _on_yes():
#             # Delete old file
#             os.remove(file_props.file_name)

#             # Request the file to the sender
#             sid = get_random_string()
#             new_file_props = FilesProp.getNewFileProp(account, sid)
#             new_file_props.file_name = file_props.file_name
#             new_file_props.name = file_props.name
#             new_file_props.desc = file_props.desc
#             new_file_props.size = file_props.size
#             new_file_props.date = file_props.date
#             new_file_props.hash_ = file_props.hash_
#             new_file_props.type_ = 'r'
#             tsid = client.get_module('Jingle').start_file_transfer(
#                 str(jid), new_file_props, True)

#             new_file_props.transport_sid = tsid
#             self.add_transfer(account, contact, new_file_props)

#         client = app.get_client(account)
#         contact = client.get_module('Contacts').get_contact(jid)

#         if file_props.type_ == 'r':
#             file_name = os.path.basename(file_props.file_name)
#         else:
#             file_name = file_props.name
#         ConfirmationAlertDialog(
#             _('File Transfer Error'),
#             _('The file %s has been received, but it seems to have '
#               'been damaged along the way.\n'
#               'Do you want to download it again?') % file_name,
#             [DialogButton.make('Cancel',
#                                text=_('_No')),
#              DialogButton.make('Accept',
#                                text=_('_Download Again'),
#                                callback=_on_yes)])

#     def send_file(self,
#                   account: str,
#                   contact: types.ResourceContact,
#                   resource_jid: JID,
#                   file_path: str,
#                   file_desc: str = ''
#                   ) -> bool:
#         '''
#         Start the real transfer(upload) of the file
#         '''
#         if file_is_locked(file_path):
#             pritext = _('File Not Readable')
#             sextext = _('Another process is using this file.')
#             InformationAlertDialog(pritext, sextext)
#             return False

#         file_name = os.path.split(file_path)[1]
#         file_props = self.get_send_file_props(
#             account,
#             resource_jid,
#             file_path,
#             file_name,
#             file_desc)
#         if file_props is None:
#             return False

#         ft_data = mod.FileTransfer(
#             state=1,
#             source=[mod.JingleFT(type='jingleft', sid=file_props.sid)],
#         )

#         message_data = mod.Message(
#             account_=account,
#             remote_jid_=contact.jid.new_as_bare(),
#             resource=resource_jid.resource,
#             type=MessageType.CHAT,
#             direction=ChatDirection.OUTGOING,
#             timestamp=utc_now(),
#             state=MessageState.ACKNOWLEDGED,
#             id=str(uuid.uuid4()),
#             filetransfers=[ft_data],
#         )

#         app.storage.archive.insert_object(message_data)

#         client = app.get_client(account)
#         client.get_module('Jingle').start_file_transfer(
#             str(resource_jid), file_props)
#         app.ged.raise_event(
#             FileRequestSent(file_props=file_props,
#                             account=account,
#                             jid=contact.jid))
#         self.add_transfer(account, contact, file_props)

#         return True

#     def _start_receive(self, file_path, account, contact, file_props):
#         file_dir = os.path.dirname(file_path)
#         if file_dir:
#             app.settings.set('last_save_dir', file_dir)
#         file_props.file_name = file_path
#         file_props.type_ = 'r'
#         self.add_transfer(account, contact, file_props)
#         client = app.get_client(account)
#         client.get_module('Bytestream').send_file_approval(file_props)

#     def on_file_request_accepted(self,
#                                  account: str,
#                                  contact: BareContact,
#                                  file_props: FileProp
#                                  ) -> None:
#         def _on_accepted(account, contact, file_props, file_paths):
#             file_path = file_paths[0]
#             if os.path.exists(file_path):
#                 app.settings.set('last_save_dir', os.path.dirname(file_path))

#                 # Check if we have write permissions
#                 if not os.access(file_path, os.W_OK):
#                     file_name = GLib.markup_escape_text(
#                         os.path.basename(file_path))
#                     InformationAlertDialog(
#                         _('File Already Exists'),
#                         _('Cannot overwrite existing file "%s". '
#                           'A file with this name already exists and you do '
#                           'not have permission to overwrite it.') % file_name
#                     )
#                     return

#             # File does not exist yet
#             dirname = os.path.dirname(file_path)
#             if not os.access(dirname, os.W_OK) and os.name != 'nt':
#                 # read-only bit is used to mark special folder under
#                 # windows, not to mark that a folder is read-only.
#                 # See ticket #3587
#                 InformationAlertDialog(
#                     _('Directory Not Writable'),
#                     _('Directory "%s" is not writable. '
#                       'You do not have permissions to create files '
#                       'in this directory.') % dirname
#                 )
#                 return
#             self._start_receive(file_path, account, contact, file_props)

#         # Show file save as dialog
#         client = app.get_client(account)
#         accept_cb = partial(_on_accepted, account, contact, file_props)
#         cancel_cb = partial(
#             client.get_module('Bytestream').send_file_rejection, file_props)
#         FileSaveDialog(accept_cb,
#                        cancel_cb,
#                        path=app.settings.get('last_save_dir'),
#                        file_name=file_props.name)

#     def set_status(self, file_props, status):
#         '''
#         Change the status of a transfer to state 'status'
#         '''
#         iter_ = self.get_iter_by_sid(file_props.type_, file_props.sid)
#         if iter_ is None:
#             return

#         if status == 'stop':
#             file_props.stopped = True
#         elif status == 'ok':
#             file_props.completed = True
#             text = self._format_percent(100)
#             received_size = GLib.format_size_full(
#                 int(file_props.received_len), self.units)
#             full_size = GLib.format_size_full(file_props.size, self.units)
#             text += received_size + '/' + full_size
#             self.model.set(iter_, Column.PROGRESS, text)
#             self.model.set(iter_, Column.PULSE, GLib.MAXINT32)
#         elif status == 'computing':
#             self.model.set(iter_, Column.PULSE, 1)
#             text = _('Checking file…') + '\n'
#             received_size = GLib.format_size_full(
#                 int(file_props.received_len), self.units)
#             full_size = GLib.format_size_full(file_props.size, self.units)
#             text += received_size + '/' + full_size
#             self.model.set(iter_, Column.PROGRESS, text)

#             def _pulse():
#                 pulse_value = self.model.get(iter_, Column.PULSE)[0]
#                 if pulse_value == GLib.MAXINT32:
#                     return False
#                 self.model.set(iter_, Column.PULSE, pulse_value + 1)
#                 return True
#             GLib.timeout_add(100, _pulse)
#         elif status == 'hash_error':
#             text = _('File error') + '\n'
#             received_size = GLib.format_size_full(
#                 int(file_props.received_len), self.units)
#             full_size = GLib.format_size_full(file_props.size, self.units)
#             text += received_size + '/' + full_size
#             self.model.set(iter_, Column.PROGRESS, text)
#             self.model.set(iter_, Column.PULSE, GLib.MAXINT32)
#         self.model.set(iter_, Column.IMAGE, self.icons[status])
#         path = self.model.get_path(iter_)
#         self._select_func(path)

#     def _format_percent(self, percent):
#         '''
#         Add extra spaces from both sides of the percent, so that progress
#         string has always a fixed size
#         '''
#         _str = '          '
#         if percent != 100.:
#             _str += ' '
#         if percent < 10:
#             _str += ' '
#         _str += str(percent) + '%          \n'
#         return _str

#     def _get_eta_and_speed(self, full_size, transferred_size, file_props):
#         if not file_props.transferred_size:
#             return 0., 0.

#         if file_props.elapsed_time == 0:
#             return 0., 0.

#         if len(file_props.transferred_size) == 1:
#             speed = round(float(transferred_size) / file_props.elapsed_time)
#         else:
#             # first and last are (time, transferred_size)
#             first = file_props.transferred_size[0]
#             last = file_props.transferred_size[-1]
#             transferred = last[1] - first[1]
#             tim = last[0] - first[0]
#             if tim == 0:
#                 return 0., 0.
#             speed = round(float(transferred) / tim)
#         if speed == 0.:
#             return 0., 0.
#         remaining_size = full_size - transferred_size
#         eta = remaining_size / speed
#         return eta, speed

#     def _remove_transfer(self, iter_, sid, file_props):
#         self.model.remove(iter_)
#         if not file_props:
#             return
#         if file_props.tt_account:
#             # file transfer is set
#             account = file_props.tt_account
#             if account in app.settings.get_active_accounts():
#                 # there is a connection to the account
#                 client = app.get_client(account)
#                 client.get_module('Bytestream').remove_transfer(file_props)

#         FilesProp.deleteFileProp(file_props)
#         del file_props

#     def _file_progress(self, event):
#         self.set_progress(
#             event.file_props.type_,
#             event.file_props.sid,
#             event.file_props.received_len)

#     def set_progress(self, typ, sid, transferred_size, iter_=None):
#         '''
#         Change the progress of a transfer with new transferred size
#         '''
#         if time.time() - self._last_progress_update < 0.5:
#             # Update window every 500ms only
#             return
#         self._last_progress_update = time.time()

#         file_props = FilesProp.getFilePropByType(typ, sid)
#         if file_props is None:
#             return

#         full_size = file_props.size
#         if full_size == 0:
#             percent = 0
#         else:
#             percent = round(float(transferred_size) / full_size * 100, 1)
#         if iter_ is None:
#             iter_ = self.get_iter_by_sid(typ, sid)
#         if iter_ is not None:
#             just_began = False
#             if self.model[iter_][Column.PERCENT] == 0 and int(percent > 0):
#                 just_began = True
#             text = self._format_percent(percent)
#             if transferred_size == 0:
#                 text += '0'
#             else:
#                 text += GLib.format_size_full(transferred_size, self.units)
#             text += '/' + GLib.format_size_full(full_size, self.units)
#             # Kb/s

#             # remaining time
#             if file_props.offset:
#                 transferred_size -= file_props.offset
#                 full_size -= file_props.offset

#             if file_props.elapsed_time > 0:
#                 file_props.transferred_size.append((file_props.last_time,
#                                                    transferred_size))
#             if len(file_props.transferred_size) > 6:
#                 file_props.transferred_size.pop(0)
#             eta, speed = self._get_eta_and_speed(full_size, transferred_size,
#                                                  file_props)

#             self.model.set(iter_, Column.PROGRESS, text)
#             self.model.set(iter_, Column.PERCENT, int(percent))
#             text = format_eta(eta)
#             text += '\n'
#             # Translators:
#             # This should make the string KB/s,
#             # where 'KB' part is taken from %s.
#             # Only the 's' after / (which means second) should be translated.
#             text += _('(%(filesize_unit)s/s)') % {
#                 'filesize_unit': GLib.format_size_full(speed, self.units)}
#             self.model.set(iter_, Column.TIME, text)

#             # try to guess what should be the status image
#             if file_props.type_ == 'r':
#                 status = 'download'
#             else:
#                 status = 'upload'
#             if file_props.paused is True:
#                 status = 'pause'
#             elif file_props.stalled is True:
#                 status = 'waiting'
#             if file_props.connected is False:
#                 status = 'stop'
#             self.model.set(iter_, 0, self.icons[status])
#             if transferred_size == full_size:
#                 # If we are receiver and this is a jingle session
#                 if (file_props.type_ == 'r' and
#                         file_props.session_type == 'jingle' and
#                         file_props.hash_):
#                     # Show that we are computing the hash
#                     self.set_status(file_props, 'computing')
#                 else:
#                     self.set_status(file_props, 'ok')
#             elif just_began:
#                 path = self.model.get_path(iter_)
#                 self._select_func(path)

#     def get_iter_by_sid(self, typ, sid):
#         '''
#         Return iter to the row, which holds file transfer, identified by the
#         session id
#         '''
#         iter_ = self.model.get_iter_first()
#         while iter_:
#             if typ + sid == self.model[iter_][Column.SID]:
#                 return iter_
#             iter_ = self.model.iter_next(iter_)

#     @staticmethod
#     def __convert_date(epoch: float) -> str:
#         # Converts date-time from seconds from epoch to iso 8601
#         dt = datetime.fromtimestamp(epoch, timezone.utc)
#         return dt.isoformat() + 'Z'

#     def get_send_file_props(self,
#                             account: str,
#                             resource_jid: JID,
#                             file_path: str,
#                             file_name: str,
#                             file_desc: str = ''
#                             ) -> FileProp | None:
#         '''
#         Create new file_props object and set initial file transfer
#         properties in it
#         '''
#         if os.path.isfile(file_path):
#             stat = os.stat(file_path)
#         else:
#             InformationAlertDialog(
#                 _('Invalid File'),
#                 _('File: %s') % file_path)
#             return None

#         if stat[6] == 0:
#             InformationAlertDialog(
#                 _('Invalid File'),
#                 _('It is not possible to send empty files'))
#             return None

#         file_props = FilesProp.getNewFileProp(
#             account,
#             sid=get_random_string())
#         mod_date = os.path.getmtime(file_path)
#         file_props.file_name = file_path
#         file_props.name = file_name
#         file_props.date = self.__convert_date(mod_date)
#         file_props.type_ = 's'
#         file_props.desc = file_desc
#         file_props.elapsed_time = 0
#         file_props.size = stat[6]
#         file_props.sender = account
#         file_props.receiver = str(resource_jid)
#         file_props.tt_account = account
#         return file_props

#     def add_transfer(self, account, contact, file_props):
#         '''
#         Add new transfer to FT window and show the FT window
#         '''
#         if file_props is None:
#             return

#         file_props.elapsed_time = 0
#         iter_ = self.model.prepend()
#         if file_props.type_ == 'r':
#             text_labels = '\n<b>' + _('Sender: ') + '</b>'
#         else:
#             text_labels = '\n<b>' + _('Recipient: ') + '</b>'

#         if file_props.type_ == 'r':
#             file_name = os.path.split(file_props.file_name)[1]
#         else:
#             file_name = file_props.name
#         text_props = GLib.markup_escape_text(file_name) + '\n'
#         text_props += contact.name
#         self.model.set(iter_,
#                        1,
#                        text_labels,
#                        2,
#                        text_props,
#                        Column.PULSE,
#                        -1,
#                        Column.SID,
#                        file_props.type_ + file_props.sid)
#         self.set_progress(file_props.type_, file_props.sid, 0, iter_)
#         if file_props.started is False:
#             status = 'waiting'
#         elif file_props.type_ == 'r':
#             status = 'download'
#         else:
#             status = 'upload'
#         file_props.tt_account = account
#         self.set_status(file_props, status)
#         self._set_cleanup_sensitivity()

#     def _on_transfers_list_row_activated(self, widget, path, col):
#         # try to open the containing folder
#         self._on_open_folder_menuitem_activate(widget)

#     def _set_cleanup_sensitivity(self):
#         '''
#         Check if there are transfer rows and set cleanup_button sensitive, or
#         insensitive if model is empty
#         '''
#         if not self.model:
#             self._ui.cleanup_button.set_sensitive(False)
#         else:
#             self._ui.cleanup_button.set_sensitive(True)

#     def _set_all_insensitive(self):
#         '''
#         Make all buttons/menuitems insensitive
#         '''
#         self._ui.pause_resume_button.set_sensitive(False)
#         self._ui.pause_resume_menuitem.set_sensitive(False)
#         self._ui.remove_menuitem.set_sensitive(False)
#         self._ui.cancel_button.set_sensitive(False)
#         self._ui.cancel_menuitem.set_sensitive(False)
#         self._ui.open_folder_menuitem.set_sensitive(False)
#         self._set_cleanup_sensitivity()

#     def _set_buttons_sensitive(self, path, is_row_selected):
#         '''
#         Make buttons/menuitems sensitive as appropriate to the state of file
#         transfer located at path 'path'
#         '''
#         if path is None:
#             self._set_all_insensitive()
#             return
#         current_iter = self.model.get_iter(path)
#         sid = self.model[current_iter][Column.SID]
#         file_props = FilesProp.getFilePropByType(sid[0], sid[1:])
#         self._ui.remove_menuitem.set_sensitive(is_row_selected)
#         self._ui.open_folder_menuitem.set_sensitive(is_row_selected)
#         is_stopped = False
#         if is_transfer_stopped(file_props):
#             is_stopped = True
#         self._ui.cancel_button.set_sensitive(not is_stopped)
#         self._ui.cancel_menuitem.set_sensitive(not is_stopped)
#         if not is_row_selected:
#             # No selection, disable the buttons
#             self._set_all_insensitive()
#         elif not is_stopped and file_props.continue_cb:
#             if is_transfer_active(file_props):
#                 # File transfer is active
#                 self._toggle_pause_continue(True)
#                 self._ui.pause_resume_button.set_sensitive(True)
#                 self._ui.pause_resume_menuitem.set_sensitive(True)
#             elif is_transfer_paused(file_props):
#                 # File transfer is paused
#                 self._toggle_pause_continue(False)
#                 self._ui.pause_resume_button.set_sensitive(True)
#                 self._ui.pause_resume_menuitem.set_sensitive(True)
#             else:
#                 self._ui.pause_resume_button.set_sensitive(False)
#                 self._ui.pause_resume_menuitem.set_sensitive(False)
#         else:
#             self._ui.pause_resume_button.set_sensitive(False)
#             self._ui.pause_resume_menuitem.set_sensitive(False)
#         return True

#     def _selection_changed(self, args):
#         '''
#         Selection has changed - change the sensitivity of the buttons/menuitems
#         '''
#         selection = args
#         selected = selection.get_selected_rows()
#         if selected[1] != []:
#             selected_path = selected[1][0]
#             self._select_func(selected_path)
#         else:
#             self._set_all_insensitive()

#     def _select_func(self, path):
#         is_selected = False
#         selected = self._ui.transfers_list.get_selection().get_selected_rows()
#         if selected[1] != []:
#             selected_path = selected[1][0]
#             if selected_path == path:
#                 is_selected = True
#         self._set_buttons_sensitive(path, is_selected)
#         self._set_cleanup_sensitivity()
#         return True

#     def _on_cleanup_button_clicked(self, widget):
#         i = len(self.model) - 1
#         while i >= 0:
#             iter_ = self.model.get_iter(i)
#             sid = self.model[iter_][Column.SID]
#             file_props = FilesProp.getFilePropByType(sid[0], sid[1:])
#             if is_transfer_stopped(file_props):
#                 self._remove_transfer(iter_, sid, file_props)
#             i -= 1
#         self._ui.transfers_list.get_selection().unselect_all()
#         self._set_all_insensitive()

#     def _toggle_pause_continue(self, status):
#         if status:
#             self._ui.pause_resume_button.set_icon_name(
#                 'media-playback-pause-symbolic')
#         else:
#             self._ui.pause_resume_button.set_icon_name(
#                 'media-playback-start-symbolic')

#     def _on_pause_resume_button_clicked(self, widget):
#         selected = self._ui.transfers_list.get_selection().get_selected()
#         if selected is None or selected[1] is None:
#             return
#         s_iter = selected[1]
#         sid = self.model[s_iter][Column.SID]
#         file_props = FilesProp.getFilePropByType(sid[0], sid[1:])
#         if is_transfer_paused(file_props):
#             file_props.last_time = time.time()
#             file_props.paused = False
#             status_types = {'r': 'download', 's': 'upload'}
#             self.set_status(file_props, status_types[sid[0]])
#             self._toggle_pause_continue(True)
#             if file_props.continue_cb:
#                 file_props.continue_cb()
#         elif is_transfer_active(file_props):
#             file_props.paused = True
#             self.set_status(file_props, 'pause')
#             # Reset that to compute speed only when we resume
#             file_props.transferred_size = []
#             self._toggle_pause_continue(False)

#     def _on_cancel_button_clicked(self, widget):
#         selected = self._ui.transfers_list.get_selection().get_selected()
#         if selected is None or selected[1] is None:
#             return
#         s_iter = selected[1]
#         sid = self.model[s_iter][Column.SID]
#         file_props = FilesProp.getFilePropByType(sid[0], sid[1:])
#         self.cancel_transfer(file_props)

#     def cancel_transfer(self, file_props: FileProp) -> None:
#         # TODO: does not cancel transfer somehow
#         account = file_props.tt_account
#         if account is None or account not in app.settings.get_active_accounts():
#             return
#         client = app.get_client(account)
#         # Check if we are in a IBB transfer
#         if file_props.direction:
#             client.get_module('IBB').send_close(file_props)
#         client.get_module('Bytestream').disconnect_transfer(file_props)
#         self.set_status(file_props, 'stop')

#     def _on_notify_ft_complete_toggled(self, widget, *args):
#         app.settings.set('notify_on_file_complete', widget.get_active())

#     def _on_file_transfers_dialog_delete_event(self, widget, event):
#         self.window.set_visible(False)
#         return True  # Do NOT destroy window

#     def _show_context_menu(self, event, iter_):
#         # change the sensitive property of the buttons and menuitems
#         if iter_:
#             path = self.model.get_path(iter_)
#             self._set_buttons_sensitive(path, True)

#         if event.type == Gdk.EventType.KEY_PRESS:
#             event_button = 0
#         else:
#             event_button = event.button

#         self._ui.file_transfers_menu.show_all()
#         self._ui.file_transfers_menu.popup(
#             None,
#             self._ui.transfers_list,
#             None,
#             None,
#             event_button,
#             event.time)

#     def _on_transfers_list_key_press_event(self, widget, event):
#         '''
#         When a key is pressed in the treeviews
#         '''
#         iter_ = None
#         try:
#             iter_ = self._ui.transfers_list.get_selection().get_selected()[1]
#         except TypeError:
#             self._ui.transfers_list.get_selection().unselect_all()

#         if iter_ is not None:
#             path = self.model.get_path(iter_)
#             self._ui.transfers_list.get_selection().select_path(path)

#         if event.keyval == Gdk.KEY_Menu:
#             self._show_context_menu(event, iter_)
#             return True

#     def _on_transfers_list_button_release_event(self, widget, event):
#         # hide tooltip, no matter the button is pressed
#         path = None
#         try:
#             path = self._ui.transfers_list.get_path_at_pos(int(event.x),
#                                                            int(event.y))[0]
#         except TypeError:
#             self._ui.transfers_list.get_selection().unselect_all()
#         if path is None:
#             self._set_all_insensitive()
#         else:
#             self._select_func(path)

#     def _on_transfers_list_button_press_event(self, widget, event):
#         # hide tooltip, no matter the button is pressed
#         path, iter_ = None, None
#         try:
#             path = self._ui.transfers_list.get_path_at_pos(int(event.x),
#                                                            int(event.y))[0]
#         except TypeError:
#             self._ui.transfers_list.get_selection().unselect_all()
#         if event.button == Gdk.BUTTON_SECONDARY:
#             if path:
#                 self._ui.transfers_list.get_selection().select_path(path)
#                 iter_ = self.model.get_iter(path)
#             self._show_context_menu(event, iter_)
#             if path:
#                 return True

#     def _on_open_folder_menuitem_activate(self, widget):
#         selected = self._ui.transfers_list.get_selection().get_selected()
#         if not selected or not selected[1]:
#             return
#         s_iter = selected[1]
#         sid = self.model[s_iter][Column.SID]
#         file_props = FilesProp.getFilePropByType(sid[0], sid[1:])
#         if not file_props.file_name:
#             return

#         open_file(Path(file_props.file_name), show_in_folder=True)

#     def _on_cancel_menuitem_activate(self, widget):
#         self._on_cancel_button_clicked(widget)

#     def _on_continue_menuitem_activate(self, widget):
#         self._on_pause_resume_button_clicked(widget)

#     def _on_pause_resume_menuitem_activate(self, widget):
#         self._on_pause_resume_button_clicked(widget)

#     def _on_remove_menuitem_activate(self, widget):
#         selected = self._ui.transfers_list.get_selection().get_selected()
#         if not selected or not selected[1]:
#             return
#         s_iter = selected[1]
#         sid = self.model[s_iter][Column.SID]
#         file_props = FilesProp.getFilePropByType(sid[0], sid[1:])
#         self._remove_transfer(s_iter, sid, file_props)
#         self._set_all_insensitive()

#     def _on_file_transfers_window_key_press_event(self, widget, event):
#         if event.keyval == Gdk.KEY_Escape:  # ESCAPE
#             self.window.set_visible(False)



# This file is part of Gajim.
#
# SPDX-License-Identifier: GPL-3.0-only

# Util module

from __future__ import annotations

from typing import Any
from typing import ParamSpec
from typing import Protocol
from typing import TypeVar

import datetime as dt
import logging
from collections.abc import Callable
from dataclasses import dataclass
from functools import partial
from functools import wraps

from nbxmpp.namespaces import Namespace
from nbxmpp.protocol import JID
from nbxmpp.protocol import Message
from nbxmpp.structs import MessageProperties
from nbxmpp.task import Task

from gajim.common import app
from gajim.common import types
from gajim.common.const import Trust

T = TypeVar("T")
P = ParamSpec("P")


def from_xs_boolean(value: str | bool) -> bool:
    if isinstance(value, bool):
        return value

    if value in ("1", "true", "True"):
        return True

    if value in ("0", "false", "False", ""):
        return False

    raise ValueError(f"Can't convert {value} to python boolean")


def to_xs_boolean(value: bool | None) -> str:
    # Convert to xs:boolean ('true', 'false')
    # from a python boolean (True, False) or None
    if value is True:
        return "true"

    if value is False:
        return "false"

    if value is None:
        return "false"

    raise ValueError(f"Can't convert {value} to xs:boolean")


def event_node(node: str) -> Callable[[Callable[P, object]], Callable[P, None]]:
    def event_node_decorator(func: Callable[P, object]) -> Callable[P, None]:
        @wraps(func)
        def func_wrapper(
            self: Any,
            _con: types.NBXMPPClient,
            _stanza: Message,
            properties: MessageProperties,
        ) -> None:
            if not properties.is_pubsub_event:
                return

            assert properties.pubsub_event is not None
            if properties.pubsub_event.node != node:
                return
            func(self, _con, _stanza, properties)  # type: ignore

        return func_wrapper  # type: ignore

    return event_node_decorator


def store_publish(func: Callable[P, T]) -> Callable[P, T | None]:
    @wraps(func)
    def func_wrapper(self: Any, *args: P.args, **kwargs: P.kwargs) -> T | None:
        if not app.account_is_connected(self._account):
            self._stored_publish = partial(func, self, *args, **kwargs)
            return None
        return func(self, *args, **kwargs)  # type: ignore

    return func_wrapper  # type: ignore


class LogAdapter(logging.LoggerAdapter[logging.Logger]):
    def process(self, msg: str, kwargs: Any) -> tuple[str, Any]:
        assert self.extra is not None
        return f"({self.extra['account']}) {msg}", kwargs


def as_task(func: Callable[P, T]) -> Callable[P, T]:
    @wraps(func)
    def func_wrapper(
        self: Any,
        *args: Any,
        timeout: int | None = None,
        callback: Callable[..., Any] | None = None,
        user_data: Any = None,
        **kwargs: Any,
    ) -> T:
        task_ = Task(func(self, *args, **kwargs))  # type: ignore
        task_.set_timeout(timeout)
        app.register_task(self, task_)
        task_.set_finalize_func(app.remove_task, id(self))
        task_.set_user_data(user_data)
        if callback is not None:
            task_.add_done_callback(callback)
        task_.start()
        return task_  # type: ignore

    return func_wrapper


def prepare_stanza(stanza: Message, plaintext: str) -> None:
    delete_nodes(stanza, "encrypted", Namespace.OMEMO_TEMP)
    delete_nodes(stanza, "body")
    stanza.setBody(plaintext)


def delete_nodes(stanza: Message, name: str, namespace: str | None = None) -> None:

    nodes = stanza.getTags(name, namespace=namespace)
    for node in nodes:
        stanza.delChild(node)


@dataclass
class PublicKeyData:
    active: bool
    address: JID
    label: str | None
    last_seen: dt.datetime | None
    fingerprint: str
    trust: Trust

    def pretty_fingerprint(self, wrap: bool = True) -> str:
        raise NotImplementedError


class CryptoModule(Protocol):
    def get_our_public_key(self) -> PublicKeyData | None: ...

    def get_public_keys(
        self, jid: JID, *, is_groupchat: bool
    ) -> list[PublicKeyData]: ...

    def set_public_key_trust(
        self, public_key_data: PublicKeyData, trust: Trust
    ) -> None: ...

    def remove_public_key(self, public_key_data: PublicKeyData) -> None: ...

    def clear_keylist(self) -> None: ...

    def compose_trust_uri(self, jid: JID) -> str: ...



# This file is part of Gajim.
#
# SPDX-License-Identifier: GPL-3.0-only

# XEP-0184: Message Delivery Receipts

from __future__ import annotations

import datetime as dt

import nbxmpp
from nbxmpp.modules.receipts import build_receipt
from nbxmpp.namespaces import Namespace
from nbxmpp.protocol import Message
from nbxmpp.structs import MessageProperties
from nbxmpp.structs import StanzaHandler

from gajim.common import app
from gajim.common import types
from gajim.common.events import ReceiptReceived
from gajim.common.modules.base import BaseModule
from gajim.common.modules.message_util import successfully_decrypted
from gajim.common.storage.archive import models as mod


class Receipts(BaseModule):
    def __init__(self, con: types.Client) -> None:
        BaseModule.__init__(self, con)

        self.handlers = [
            StanzaHandler(
                name="message",
                callback=self._process_message_receipt,
                ns=Namespace.RECEIPTS,
                priority=46,
            ),
        ]

    def _process_message_receipt(
        self, _con: types.NBXMPPClient, stanza: Message, properties: MessageProperties
    ) -> None:

        if not properties.is_receipt:
            return

        assert properties.receipt is not None
        if properties.type.is_error:
            if properties.receipt.is_request:
                return
            # Don't propagate this event further
            raise nbxmpp.NodeProcessed

        if (
            properties.type.is_groupchat
            or properties.is_self_message
            or (properties.carbon is not None and properties.carbon.is_sent)
        ):
            if properties.receipt.is_received:
                # Don't propagate this event further
                raise nbxmpp.NodeProcessed
            return

        if properties.receipt.is_request and not properties.is_mam_message:
            if not app.settings.get_account_setting(self._account, "answer_receipts"):
                return

            if properties.eme is not None:
                # Don't send receipt for message which couldn't be decrypted
                if not properties.is_encrypted:
                    return

            if not self._should_answer(properties):
                return
            self._log.info("Send receipt: %s", properties.jid)
            self._con.connection.send(build_receipt(stanza))
            return

        if properties.receipt.is_received:
            self._log.info("Receipt from %s %s", properties.jid, properties.receipt.id)

            if properties.mam is not None:
                timestamp = properties.mam.timestamp
            else:
                timestamp = properties.timestamp

            timestamp = dt.datetime.fromtimestamp(timestamp, dt.UTC)

            assert properties.remote_jid is not None
            assert properties.receipt.id is not None
            receipt_data = mod.Receipt(
                account_=self._account,
                remote_jid_=properties.remote_jid,
                id=properties.receipt.id,
                timestamp=timestamp,
            )
            app.storage.archive.insert_object(receipt_data)

            app.ged.raise_event(
                ReceiptReceived(
                    account=self._account,
                    jid=properties.remote_jid,
                    receipt_id=properties.receipt.id,
                )
            )

            raise nbxmpp.NodeProcessed

    def _should_answer(self, properties: MessageProperties) -> bool:
        if not successfully_decrypted(properties):
            return False

        if properties.is_muc_pm:
            return True

        assert properties.jid is not None

        item = self._con.get_module("Roster").get_item(properties.jid.new_as_bare())
        if item is None:
            return False

        return item.subscription in ("from", "both")



# Copyright (C) 2018 Philipp Hörist <philipp AT hoerist.com>
#
# This file is part of Gajim.
#
# SPDX-License-Identifier: GPL-3.0-or-later

from __future__ import annotations

from typing import cast
from typing import Literal
from typing import NamedTuple

import logging
from collections.abc import Callable
from pathlib import Path

from gi.repository import Gio
from gi.repository import GLib
from gi.repository import GObject
from gi.repository import Gtk
from gi.repository import Pango

from gajim.common import app
from gajim.common.i18n import _

from gajim.gtk.util.classes import SignalManager

AcceptCallbackT = Callable[[list[str]], None]


log = logging.getLogger("gajim.gtk.filechoosers")


class Filter(NamedTuple):
    name: str
    patterns: list[str] | None = None
    suffixes: list[str] | None = None
    default: bool = False


class FileChooserButton(Gtk.Button, SignalManager):
    _cls_filters: list[Filter] = []
    __gsignals__ = {
        "path-picked": (GObject.SignalFlags.RUN_LAST, None, (object,)),
    }

    def __init__(
        self,
        path: Path | None = None,
        mode: Literal["file-open", "folder-open", "save"] = "file-open",
        multiple: bool = False,
        filters: list[Filter] | None = None,
        label: str = "",
        tooltip: str = "",
        icon_name: str | None = None,
        initial_path: Path | None = None,
    ) -> None:
        Gtk.Button.__init__(self)
        SignalManager.__init__(self)
        self._path = path
        self._initial_path = initial_path
        self._mode = mode
        self._multiple = multiple
        self._filters = filters or self._cls_filters
        self._label_text = label
        self._has_tooltip = bool(tooltip)

        self.set_tooltip_text(tooltip)

        icon = Gtk.Image.new_from_icon_name(icon_name or "lucide-folder-symbolic")

        self._label = Gtk.Label(
            label=label,
            ellipsize=Pango.EllipsizeMode.MIDDLE,
            max_width_chars=18,
            visible=bool(label),
        )

        box = Gtk.Box(spacing=12)
        box.append(icon)
        box.append(self._label)

        if path is not None:
            self.set_path(path)

        self.set_child(box)

        self._connect(self, "clicked", self._on_clicked)

    def get_path(self) -> Path | None:
        return self._path

    def set_path(self, path: Path | None) -> None:
        if path is None:
            self.reset()
            return

        self._path = path
        if not self._has_tooltip and self._mode != "save":
            self._label.set_tooltip_text(str(path))

        match self._mode:
            case "file-open":
                self._label.set_text(_("File: %s") % path.name)
            case "folder-open":
                self._label.set_text(_("Folder: %s") % path.as_posix())
            case _:
                pass

    def set_inital_path(self, path: Path | None) -> None:
        self._initial_path = path

    def reset(self) -> None:
        self._path = None
        self._label.set_text(self._label_text or "")
        self._label.set_tooltip_text(None)

    def _on_clicked(self, _button: FileChooserButton) -> None:
        dialog = Gtk.FileDialog()

        file_filter_model = Gio.ListStore(item_type=Gtk.FileFilter)
        for f in self._filters:
            file_filter = Gtk.FileFilter(
                name=f.name, patterns=f.patterns or [], suffixes=f.suffixes or []
            )
            file_filter_model.append(file_filter)
            if f.default:
                dialog.set_default_filter(file_filter)

        if self._path is None:
            initial_path = self._initial_path
        else:
            initial_path = self._path

        if initial_path is not None:
            file = Gio.File.new_for_path(str(initial_path))
            if initial_path.is_dir():
                dialog.set_initial_folder(file)
            else:
                dialog.set_initial_file(file)

        if self._filters:
            dialog.set_filters(file_filter_model)

        parent = cast(Gtk.Window, self.get_root())

        match (self._mode, self._multiple):
            case ("file-open", True):
                dialog.open_multiple(parent, None, self._on_file_picked)
            case ("file-open", False):
                dialog.open(parent, None, self._on_file_picked)
            case ("folder-open", True):
                dialog.select_multiple_folders(parent, None, self._on_file_picked)
            case ("folder-open", False):
                dialog.select_folder(parent, None, self._on_file_picked)
            case ("save", False):
                dialog.save(parent, None, self._on_file_picked)
            case _:
                raise ValueError("Unexpected file chooser configuration")

    def _set_error(self) -> None:
        log.warning("Could not get picked file/folder")
        self._label.set_text(_("Error"))
        if not self._has_tooltip:
            self._label.set_tooltip_text(_("Could not select file or folder"))
        self.emit("path-picked", [])

    def _on_file_picked(
        self, file_dialog: Gtk.FileDialog, result: Gio.AsyncResult
    ) -> None:
        try:
            match (self._mode, self._multiple):
                case ("file-open", True):
                    g_files = cast(
                        list[Gio.File], list(file_dialog.open_multiple_finish(result))
                    )
                case ("file-open", False):
                    g_files = [file_dialog.open_finish(result)]
                case ("folder-open", True):
                    g_files = cast(
                        list[Gio.File],
                        list(file_dialog.select_multiple_folders_finish(result)),
                    )
                case ("folder-open", False):
                    g_files = [file_dialog.select_folder_finish(result)]
                case ("save", False):
                    g_files = [file_dialog.save_finish(result)]
                case _:
                    raise ValueError("Unexpected file chooser configuration")

        except GLib.Error as e:
            if e.code == 2:
                # User dismissed dialog, do nothing
                return

            log.exception(e)
            self._set_error()
            return

        paths: list[Path] = []
        for f in g_files:
            path = f.get_path()
            if path is not None:
                paths.append(Path(path))

        if len(paths) == 1:
            self.set_path(paths[0])

        self.emit("path-picked", paths)

    def do_unroot(self) -> None:
        Gtk.Button.do_unroot(self)
        self._disconnect_all()
        app.check_finalize(self)


class AvatarFileChooserButton(FileChooserButton):
    _cls_filters = [
        Filter(name=_("PNG files"), suffixes=["png"], default=True),
        Filter(name=_("JPEG files"), suffixes=["jpg", "jpeg"]),
        Filter(name=_("SVG files"), suffixes=["svg"]),
    ]



# This file is part of Gajim.
#
# SPDX-License-Identifier: GPL-3.0-only

from typing import Any

import functools
import logging
import sys
import xml.etree.ElementTree as ET

from gi.repository import Gio
from gi.repository import Gtk

from gajim.common import configpaths
from gajim.common import i18n
from gajim.common.i18n import _

log = logging.getLogger("gajim.gtk")


class GajimBuilder:
    filename = ""

    def __init__(
        self,
        filename: str | None = None,
        instance: Any = None,
        widgets: list[str] | None = None,
        domain: str | None = None,
        gettext_: Any | None = None,
    ) -> None:
        if filename is None:
            filename = self.filename

        self._builder = Gtk.Builder(instance)

        if domain is None:
            domain = i18n.DOMAIN
        self._builder.set_translation_domain(domain)

        if gettext_ is None:
            gettext_ = _

        xml_text = self._load_string_from_filename(filename, gettext_)

        log.debug("Load ui file: %s", filename)

        if widgets is not None:
            self._builder.add_objects_from_string(xml_text, widgets)
        else:
            self._builder.add_from_string(xml_text)

    @staticmethod
    @functools.cache
    def _load_string_from_filename(filename: str, gettext_: Any) -> str:
        file_path = str(configpaths.get("GUI") / filename)

        if sys.platform == "win32":
            # This is a workaround for non working translation on Windows
            tree = ET.parse(file_path)
            for node in tree.findall(".//*[@translatable='yes']"):
                node.text = gettext_(node.text) if node.text else ""
                del node.attrib["translatable"]

            return ET.tostring(tree.getroot(), encoding="unicode", method="xml")

        file = Gio.File.new_for_path(file_path)
        content = file.load_contents(None)
        return content[1].decode()

    def __getattr__(self, name: str) -> Any:
        try:
            return getattr(self._builder, name)
        except AttributeError:
            return self._builder.get_object(name)

    def get(self, name: str) -> Any:
        return self._builder.get_object(name)


def get_builder(
    file_name: str,
    instance: Any = None,
    widgets: list[str] | None = None,
) -> GajimBuilder:
    return GajimBuilder(file_name, instance, widgets)



# This file is part of Gajim.
#
# SPDX-License-Identifier: GPL-3.0-only

from __future__ import annotations

from gi.repository import Gtk
from gi.repository import Pango
from nbxmpp.structs import VoiceRequest

from gajim.common import app
from gajim.common.i18n import _
from gajim.common.modules.contacts import GroupchatContact
from gajim.common.types import ChatContactT

from gajim.gtk.util.classes import SignalManager


class VoiceRequestsButton(Gtk.Button, SignalManager):
    def __init__(self) -> None:
        Gtk.Button.__init__(
            self,
            tooltip_text=_("Pending Voice Requests"),
            visible=False,
            halign=Gtk.Align.END,
            valign=Gtk.Align.CENTER,
        )
        SignalManager.__init__(self)

        inner_box = Gtk.Box()
        self.set_child(inner_box)

        image = Gtk.Image.new_from_icon_name("lucide-circle-question-mark-symbolic")
        inner_box.append(image)

        self.add_css_class("pulse-opacity")
        self.add_css_class("suggested-action")

        self._popover = Gtk.Popover(width_request=400, height_request=400)
        self._popover.add_css_class("p-6")

        inner_box.append(self._popover)

        self._connect(self, "clicked", self._on_button_clicked)

    def switch_contact(self, contact: ChatContactT) -> None:
        if not isinstance(contact, GroupchatContact):
            self._contact = None
            self.set_visible(False)
            self.set_visible(False)
            return

        self._contact = contact
        self._update()

    def do_unroot(self) -> None:
        Gtk.Button.do_unroot(self)
        self._disconnect_all()
        del self._popover
        app.check_finalize(self)

    def _update(self) -> None:
        self.set_visible(False)
        self.set_visible(False)

        assert self._contact is not None
        client = app.get_client(self._contact.account)
        voice_requests = client.get_module("MUC").get_voice_requests(self._contact)
        if voice_requests:
            self.set_visible(True)

    def _on_button_clicked(self, _button: VoiceRequestsButton) -> None:
        assert self._contact is not None
        client = app.get_client(self._contact.account)
        voice_requests = client.get_module("MUC").get_voice_requests(self._contact)
        if not voice_requests:
            return

        self._update_content()
        self._popover.popup()

    def _update_content(self) -> None:
        assert self._contact is not None
        client = app.get_client(self._contact.account)
        voice_requests = client.get_module("MUC").get_voice_requests(self._contact)

        menu_box = Gtk.Box(orientation=Gtk.Orientation.VERTICAL, spacing=6)
        menu_box.add_css_class("p-12")
        menu_box.set_hexpand(True)

        desc_label = Gtk.Label(
            label=_("Participants asking for voice:"),
            max_width_chars=35,
            wrap=True,
            margin_bottom=6,
        )
        desc_label.add_css_class("dimmed")
        menu_box.append(desc_label)

        scrolled = Gtk.ScrolledWindow(
            hscrollbar_policy=Gtk.PolicyType.NEVER, hexpand=True, vexpand=True
        )
        menu_box.append(scrolled)

        requests_box = Gtk.Box(orientation=Gtk.Orientation.VERTICAL, spacing=6)
        scrolled.set_child(requests_box)

        for request in voice_requests:
            request_box = Gtk.Box(spacing=12)
            requests_box.append(request_box)

            name_label = Gtk.Label(
                label=f"{request.nick} ({request.jid.bare})",
                hexpand=True,
                max_width_chars=30,
                ellipsize=Pango.EllipsizeMode.MIDDLE,
                xalign=0,
                tooltip_text=str(request.jid.bare),
            )
            request_box.append(name_label)

            decline_button = Gtk.Button.new_from_icon_name("lucide-circle-x-symbolic")
            decline_button.set_tooltip_text(_("Decline"))
            self._connect(
                decline_button, "clicked", self._on_decline, request, self._contact
            )
            request_box.append(decline_button)

            approve_button = Gtk.Button.new_from_icon_name("lucide-check-symbolic")
            approve_button.set_tooltip_text(_("Approve"))
            self._connect(
                approve_button, "clicked", self._on_approve, request, self._contact
            )
            request_box.append(approve_button)

        self._popover.set_child(menu_box)

    def _on_approve(
        self,
        _button: Gtk.Button,
        voice_request: VoiceRequest,
        contact: GroupchatContact,
    ) -> None:
        client = app.get_client(contact.account)
        client.get_module("MUC").approve_voice_request(contact, voice_request)
        self._update()
        self._update_content()

    def _on_decline(
        self,
        _button: Gtk.Button,
        voice_request: VoiceRequest,
        contact: GroupchatContact,
    ) -> None:
        client = app.get_client(contact.account)
        client.get_module("MUC").decline_voice_request(contact, voice_request)
        self._update()
        self._update_content()



# This file is part of Gajim.
#
# SPDX-License-Identifier: GPL-3.0-only

from __future__ import annotations

import locale
import logging
from pathlib import Path

from gi.repository import Adw
from gi.repository import Gdk
from gi.repository import Gio
from gi.repository import GLib
from gi.repository import GObject
from gi.repository import Gtk
from packaging.version import Version

from gajim.common import app
from gajim.common import configpaths
from gajim.common import ged
from gajim.common.exceptions import PluginsystemError
from gajim.common.i18n import _
from gajim.common.types import PluginRepositoryT
from gajim.plugins.events import PluginAdded
from gajim.plugins.events import PluginRemoved
from gajim.plugins.helpers import GajimPluginActivateException
from gajim.plugins.manifest import PluginManifest
from gajim.plugins.repository import PluginRepository

from gajim.gtk.alert import ConfirmationAlertDialog
from gajim.gtk.alert import InformationAlertDialog
from gajim.gtk.filechoosers import FileChooserButton
from gajim.gtk.filechoosers import Filter
from gajim.gtk.settings import GajimPreferencesGroup
from gajim.gtk.util.classes import SignalManager
from gajim.gtk.util.misc import get_ui_string
from gajim.gtk.util.window import get_app_window

log = logging.getLogger("gajim.gtk.plugins")


class Plugins(GajimPreferencesGroup):
    def __init__(self) -> None:
        GajimPreferencesGroup.__init__(self, key="plugins", title=_("Plugins"))

        self._plugin_rows: dict[str, PluginRow] = {}

        self._file_chooser_button = FileChooserButton(
            filters=[
                Filter(name=_("All files"), patterns=["*"]),
                Filter(name=_("ZIP files"), suffixes=["zip"], default=True),
            ],
            label=_("Install from Archive"),
            tooltip=_("Install Plugin from ZIP-File"),
            icon_name="lucide-package-plus-symbolic",
        )
        self._file_chooser_button.set_visible(not app.is_flatpak())
        self._file_chooser_button.set_valign(Gtk.Align.CENTER)
        self._connect(
            self._file_chooser_button, "path-picked", self._on_install_plugin_from_zip
        )

        description_text = _("Manage and configure Gajim plugins")
        if app.is_flatpak():
            flatpak_howto_text = _("How to install plugins with Flatpak")
            flatpak_howto_url = "https://dev.gajim.org/gajim/gajim/wikis/help/flathub"
            description_text += (
                f"\n<a href='{flatpak_howto_url}'>{flatpak_howto_text}</a>"
            )

        self.set_description(description_text)
        self.set_header_suffix(self._file_chooser_button)

        self.register_events(
            [
                ("plugin-removed", ged.GUI1, self._on_plugin_removed),
                ("plugin-added", ged.GUI1, self._on_plugin_added),
            ]
        )

        app.plugin_repository.connect("download-started", self._on_download_started)
        app.plugin_repository.connect("download-finished", self._on_download_finished)
        app.plugin_repository.connect("download-failed", self._on_download_failed)
        app.plugin_repository.connect(
            "plugin-updates-available", self._on_plugin_updates_available
        )

        # Load manifests in idle fashion, since it blocks the UI
        GLib.idle_add(self._load_installed_manifests)
        GLib.idle_add(self._load_repository_manifests)

    def do_unroot(self) -> None:
        app.plugin_repository.disconnect(self)
        GajimPreferencesGroup.do_unroot(self)

    @staticmethod
    def _sort_func(row1: PluginRow, row2: PluginRow) -> int:
        return locale.strcoll(row1.name, row2.name)

    def _load_installed_manifests(self) -> None:
        for plugin in sorted(app.plugin_manager.plugins, key=lambda p: p.manifest.name):
            self._add_manifest(plugin.manifest, True)

    def _load_repository_manifests(self) -> None:
        for manifest in app.plugin_repository.get_manifests():
            self._add_manifest(manifest, False)

    def _get_restart(self, manifest: PluginManifest) -> bool:
        path = configpaths.get("PLUGINS_DOWNLOAD") / manifest.short_name
        return path.exists()

    def _add_manifest(self, manifest: PluginManifest, installed: bool) -> None:
        restart = self._get_restart(manifest)

        plugin_row = self._plugin_rows.get(manifest.short_name)
        if plugin_row is None:
            plugin_row = PluginRow(manifest, installed)
            self._plugin_rows[manifest.short_name] = plugin_row
            self.add(plugin_row)

        elif restart:
            plugin_row.set_requires_restart(True)

        else:
            if manifest.version > plugin_row.version:
                plugin_row.set_update_available(True, manifest.version)

    def _on_plugin_added(self, event: PluginAdded) -> None:
        self._add_manifest(event.manifest, True)

    def _on_plugin_removed(self, event: PluginRemoved) -> None:
        row = self._plugin_rows[event.manifest.short_name]

        if not app.plugin_repository.contains(event.manifest.short_name):
            self.remove(row)
            return

        row.set_requires_restart(False)
        row.set_update_available(False)
        row.set_installed(False)

    def _on_install_plugin_from_zip(
        self, _button: FileChooserButton, paths: list[Path]
    ) -> None:
        if not paths:
            return

        zip_filename = str(paths[0])

        def _on_response() -> None:
            plugin = app.plugin_manager.install_from_zip(zip_filename, overwrite=True)
            if not plugin:
                InformationAlertDialog(
                    _("Archive Malformed"), _("Archive is malformed")
                )
                return

        try:
            plugin = app.plugin_manager.install_from_zip(zip_filename)
        except PluginsystemError as er_type:
            error_text = str(er_type)
            if error_text == _("Plugin already exists"):
                ConfirmationAlertDialog(
                    _("Overwrite Plugin?"),
                    _("Do you want to overwrite the currently installed version?"),
                    confirm_label=_("_Overwrite"),
                    appearance="destructive",
                    callback=_on_response,
                )
                return

            InformationAlertDialog(error_text, f'"{zip_filename}"')
            return

        if not plugin:
            InformationAlertDialog(_("Archive Malformed"), _("Archive is malformed"))

    def _on_download_started(
        self,
        _repository: PluginRepositoryT,
        _signal_name: str,
        manifests: set[PluginManifest],
    ) -> None:
        for manifest in manifests:
            row = self._plugin_rows[manifest.short_name]
            row.set_update_available(False)
            row.set_error(False)
            row.set_downloading(True)

    def _on_download_failed(
        self,
        _repository: PluginRepositoryT,
        _signal_name: str,
        manifest: PluginManifest,
        error: str,
    ) -> None:
        row = self._plugin_rows[manifest.short_name]
        assert row is not None
        row.set_downloading(False)
        row.set_error(True, error)

    def _on_download_finished(
        self,
        _repository: PluginRepositoryT,
        _signal_name: str,
        manifest: PluginManifest,
    ) -> None:
        row = self._plugin_rows[manifest.short_name]
        assert row is not None
        row.set_downloading(False)
        row.update_manifest(manifest)

        activated = app.plugin_manager.update_plugins(
            replace=False, activate=True, plugin_name=manifest.short_name
        )
        if activated:
            row.set_installed(True)
        else:
            row.set_requires_restart(True)
            log.info("Plugin %s needs restart", manifest.short_name)

    def _on_plugin_updates_available(
        self,
        _repository: PluginRepository,
        _signal_name: str,
        manifests: list[PluginManifest],
    ) -> None:
        for manifest in manifests:
            row = self._plugin_rows[manifest.short_name]
            row.set_update_available(True, new_version=manifest.version)


@Gtk.Template(string=get_ui_string("plugin_row.ui"))
class PluginRow(Adw.ExpanderRow, SignalManager):
    __gtype_name__ = "PluginRow"

    _warning_icon: Gtk.Image = Gtk.Template.Child()
    _update_badge: Gtk.Label = Gtk.Template.Child()
    _requires_restart_badge: Gtk.Label = Gtk.Template.Child()
    _downloading_spinner: Adw.Spinner = Gtk.Template.Child()
    _config_button: Gtk.Button = Gtk.Template.Child()
    _enable_switch: Gtk.Switch = Gtk.Template.Child()
    _install_button: Gtk.Button = Gtk.Template.Child()
    _authors_row: Adw.ActionRow = Gtk.Template.Child()
    _url_row: Adw.ActionRow = Gtk.Template.Child()
    _version_row: Adw.ActionRow = Gtk.Template.Child()
    _warning_row: Adw.ActionRow = Gtk.Template.Child()
    _management_row: Adw.ActionRow = Gtk.Template.Child()
    _update_button: Gtk.Button = Gtk.Template.Child()
    _uninstall_button: Gtk.Button = Gtk.Template.Child()

    def __init__(self, manifest: PluginManifest, installed: bool) -> None:
        Adw.ExpanderRow.__init__(
            self, title=manifest.name, subtitle=manifest.description
        )
        SignalManager.__init__(self)

        self._manifest = manifest
        self._installed = installed

        self.add_prefix(self._get_plugin_icon())

        plugin = app.plugin_manager.get_plugin(manifest.short_name)
        plugin_active = plugin.active if plugin is not None else False

        self._enable_switch.set_active(plugin_active)
        self._connect(self._config_button, "clicked", self._on_config_clicked)
        self._active_signal_id = self._connect(
            self._enable_switch, "notify::active", self._on_switch_active
        )
        self._connect(self._install_button, "clicked", self._on_install_clicked)
        self._connect(self._update_button, "clicked", self._on_install_clicked)
        self._connect(self._uninstall_button, "clicked", self._on_uninstall_clicked)

        self._update()

    def do_unroot(self) -> None:
        self._disconnect_all()
        Adw.ExpanderRow.do_unroot(self)

    @property
    def name(self) -> str:
        return self._manifest.name

    @property
    def version(self) -> Version:
        return self._manifest.version

    def set_installed(self, installed: bool) -> None:
        self._installed = installed
        self._update()

    def set_error(self, has_error: bool, error: str | None = None) -> None:
        self._warning_icon.set_visible(has_error)
        self._warning_icon.set_tooltip_text(error or "")

        self._warning_row.set_visible(has_error)
        self._warning_row.set_subtitle(error or "")

    def set_downloading(self, downloading: bool) -> None:
        self._downloading_spinner.set_visible(downloading)

    def set_requires_restart(self, requires_restart: bool) -> None:
        self._requires_restart_badge.set_visible(requires_restart)
        self._install_button.set_sensitive(not requires_restart)

    def set_update_available(
        self, update_available: bool, new_version: Version | None = None
    ) -> None:
        self._update_badge.set_visible(update_available)
        self._update_button.set_visible(update_available)
        self._update_button.set_sensitive(True)

        if update_available:
            self._management_row.set_subtitle(
                _("Update available: Version %s") % str(new_version)
            )
        else:
            self._management_row.set_subtitle("")

    def update_manifest(self, manifest: PluginManifest) -> None:
        self._manifest = manifest
        self._update()

    def _update(self) -> None:
        authors = [GLib.markup_escape_text(author) for author in self._manifest.authors]
        self._authors_row.set_subtitle("\n".join(authors))
        self._url_row.set_subtitle(
            f"<a href='{self._manifest.homepage}'>{self._manifest.homepage}</a>"
        )
        self._version_row.set_subtitle(str(self._manifest.version))

        has_error, error_text = self._get_error()
        self._warning_icon.set_visible(has_error)
        self._warning_icon.set_tooltip_text(error_text)

        self._warning_row.set_visible(has_error)
        self._warning_row.set_subtitle(error_text)

        plugin = app.plugin_manager.get_plugin(self._manifest.short_name)
        plugin_active = plugin.active if plugin is not None else False
        plugin_activatable = plugin.activatable if plugin is not None else False

        self._config_button.set_visible(
            self._installed and self._manifest.config_dialog and not has_error
        )
        self._config_button.set_sensitive(plugin_active)

        self._enable_switch.set_visible(self._installed and not has_error)
        self._enable_switch.set_sensitive(plugin_activatable)
        self._enable_switch.set_active(plugin_active)

        self._install_button.set_visible(not self._installed and not app.is_flatpak())
        self._install_button.set_sensitive(not self._installed)

        self._management_row.set_visible(self._installed and not app.is_flatpak())

        if self._manifest.is_shipped:
            self._management_row.set_subtitle(
                _("This plugin is managed by your operating system")
            )
            self._uninstall_button.set_sensitive(False)

    def _get_plugin_icon(self) -> Gtk.Image:
        image = Gtk.Image.new_from_gicon(Gio.ThemedIcon(name="lucide-package-symbolic"))

        image_name = f"{self._manifest.short_name}.png"
        path = configpaths.get("PLUGINS_IMAGES") / image_name
        if path.exists():
            texture = Gdk.Texture.new_from_filename(str(path))
            image.set_from_paintable(texture)
            return image

        plugin = app.plugin_manager.get_plugin(self._manifest.short_name)
        if plugin is None:
            return image

        path = Path(plugin.__path__) / image_name
        if path.exists():
            texture = Gdk.Texture.new_from_filename(str(path))
            image.set_from_paintable(texture)
            return image

        image.add_css_class("dimmed")
        return image

    def _get_error(self) -> tuple[bool, str]:
        """Returns has_error, error_text"""
        if not self._installed:
            return False, ""

        plugin = app.plugin_manager.get_plugin(self._manifest.short_name)
        if plugin is None:
            return True, _("Plugin could not be loaded")

        if not plugin.activatable:
            return True, plugin.available_text
        return False, ""

    def _on_switch_active(self, switch: Gtk.Switch, spec: GObject.ParamSpec) -> None:
        plugin = app.plugin_manager.get_plugin(self._manifest.short_name)
        assert plugin is not None

        if switch.get_active():
            try:
                app.plugin_manager.activate_plugin(plugin)
            except GajimPluginActivateException as e:
                self._config_button.set_sensitive(False)
                with switch.handler_block(self._active_signal_id):
                    switch.set_active(False)
                InformationAlertDialog(_("Plugin Failed"), str(e))

            else:
                self._config_button.set_sensitive(True)

        else:
            app.plugin_manager.deactivate_plugin(plugin)
            self._config_button.set_sensitive(False)

    def _on_config_clicked(self, _button: Gtk.Button) -> None:
        plugin = app.plugin_manager.get_plugin(self._manifest.short_name)
        assert plugin is not None
        plugin.config_dialog(get_app_window("Preferences"))  # type: ignore

    def _on_install_clicked(self, button: Gtk.Button) -> None:
        button.set_sensitive(False)
        app.plugin_repository.download_plugins([self._manifest])

    def _on_uninstall_clicked(self, _button: Gtk.Button) -> None:
        plugin = app.plugin_manager.get_plugin(self._manifest.short_name)
        error_text = _("Unable to properly remove the plugin")
        if plugin is None:
            InformationAlertDialog(_("Warning"), error_text)
            return

        try:
            app.plugin_manager.uninstall_plugin(plugin)
        except PluginsystemError as error:
            InformationAlertDialog(_("Warning"), f"{error_text}\n{error}")
            return



# This file is part of Gajim.
#
# SPDX-License-Identifier: GPL-3.0-only

# All XEPs that don’t need their own module


import logging

from nbxmpp.structs import MessageProperties

from gajim.common.storage.archive import models as mod

log = logging.getLogger("gajim.c.m.misc")


# XEP-0066: Out of Band Data
def parse_oob(properties: MessageProperties) -> list[mod.OOB]:
    if not properties.is_oob:
        return []

    assert properties.oob is not None

    return [mod.OOB(url=properties.oob.url, description=properties.oob.desc)]



# This file is part of Gajim.
#
# SPDX-License-Identifier: GPL-3.0-only

# XEP-0118: User Tune

from __future__ import annotations

from typing import Any

from nbxmpp.namespaces import Namespace
from nbxmpp.protocol import JID
from nbxmpp.structs import MessageProperties
from nbxmpp.structs import TuneData

from gajim.common import app
from gajim.common import ged
from gajim.common import types
from gajim.common.dbus.music_track import MusicTrackListener
from gajim.common.events import MusicTrackChanged
from gajim.common.events import SignedIn
from gajim.common.modules.base import BaseModule
from gajim.common.modules.util import event_node
from gajim.common.modules.util import store_publish
from gajim.common.util.decorators import event_filter


class UserTune(BaseModule):
    _nbxmpp_extends = "Tune"
    _nbxmpp_methods = [
        "set_tune",
    ]

    def __init__(self, con: types.Client) -> None:
        BaseModule.__init__(self, con)
        self._register_pubsub_handler(self._tune_received)
        self._current_tune: TuneData | None = None
        self._contact_tunes: dict[JID, TuneData | None] = {}

    def get_current_tune(self) -> TuneData | None:
        return self._current_tune

    def get_contact_tune(self, jid: JID) -> TuneData | None:
        return self._contact_tunes.get(jid)

    @event_node(Namespace.TUNE)
    def _tune_received(
        self, _con: types.NBXMPPClient, _stanza: Any, properties: MessageProperties
    ) -> None:
        assert properties.pubsub_event is not None
        if properties.pubsub_event.retracted:
            return

        data = properties.pubsub_event.data
        if properties.is_self_message:
            self._current_tune = data

        assert properties.jid is not None
        self._contact_tunes[properties.jid] = data

        contact = self._get_contact(properties.jid)
        contact.notify("tune-update", data)

    @store_publish
    def set_tune(self, tune: TuneData | None, force: bool = False) -> None:
        if not self._con.get_module("PEP").supported:
            return

        if not force and not app.settings.get_account_setting(
            self._account, "publish_tune"
        ):
            return

        if tune == self._current_tune:
            return

        self._current_tune = tune

        self._log.info("Send %s", tune)
        self._nbxmpp("Tune").set_tune(tune)

    def set_enabled(self, enable: bool) -> None:
        if enable:
            self.register_events(
                [
                    ("music-track-changed", ged.CORE, self._on_music_track_changed),
                    ("signed-in", ged.CORE, self._on_signed_in),
                ]
            )
            self._publish_current_tune()
        else:
            self.unregister_events()
            self.set_tune(None, force=True)

    def _publish_current_tune(self):
        self.set_tune(MusicTrackListener.get().current_tune)

    @event_filter(["account"])
    def _on_signed_in(self, _event: SignedIn) -> None:
        self._publish_current_tune()

    def _on_music_track_changed(self, event: MusicTrackChanged) -> None:
        if self._current_tune == event.info:
            return

        self.set_tune(event.info)
